# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

import ctypes
import functools
import os
import sys
from io import StringIO
from pathlib import Path

from numba_cuda_mlir._mlir.passmanager import PassManager
from numba_cuda_mlir._threading import _LockedCounter
from numba_cuda_mlir._mlir.dialects import llvm
from numba_cuda_mlir.tools import generate_mangled_name
from numba_cuda_mlir._mlir import ir
from numba_cuda_mlir.lowering_utilities import context
from numba_cuda_mlir.optimization import run_pre_codegen_patterns
from numba_cuda_mlir.numba_cuda.cudadrv.nvvm import CompilationUnit
from numba_cuda_mlir.logging import trace
from numba_cuda_mlir.mlir.util import find_ops
from numba_cuda_mlir.lowering_utilities.llvm_utils import (
    LLVM_C_LIB_PATH,
    NVPTX64_DATALAYOUT,
    NVPTX64_TRIPLE,
    translate_to_llvmir,
    translate_to_llvmir_text,
    translate_gpu_module_to_libnvvm_ir,
    dump_llvmir,
)
from numba_cuda_mlir.memory_management.rtsys import rtsys


def _nrt_memsys_setup_callback(loaded_module):
    """Post-load callback that sets the shared NRT_MemSys pointer for a kernel module.

    Invoked by ``_make_post_load_hook`` after the CUlibrary is loaded.
    This ensures the kernel's ``TheMSys`` device global points to the
    shared allocation so NRT statistics are collected.
    """
    rtsys.ensure_allocated()
    rtsys.ensure_initialized()
    rtsys.set_memsys_to_library(loaded_module.handle)


def get_inline_pipeline():
    return "inline{default-pipeline=canonicalize max-iterations=10}"


def get_base_pipeline():
    inline = get_inline_pipeline()
    return (
        """
    builtin.module(
      reconcile-unrealized-casts,
      convert-shape-to-std,
      one-shot-bufferize,
      """
        + inline
        + """,
      convert-linalg-to-parallel-loops,
      convert-complex-to-standard,
      convert-complex-to-llvm,
      convert-nvgpu-to-nvvm,
      """
        + inline
        + """,
      convert-math-to-nvvm,
      gpu-kernel-outlining{data-layout-str=},
      convert-vector-to-scf{full-unroll=false lower-scalable=false lower-tensors=false target-rank=1},
      convert-scf-to-cf,
      convert-nvvm-to-llvm,
      convert-func-to-llvm{index-bitwidth=0 use-bare-ptr-memref-call-conv=false},
      expand-strided-metadata,
      lower-affine,
      math-uplift-to-fma,
      gpu.module(
        convert-gpu-to-nvvm{ has-redux=false index-bitwidth=64 use-bare-ptr-memref-call-conv=false}
      ),
      convert-arith-to-llvm{index-bitwidth=0},
      convert-index-to-llvm{index-bitwidth=64},
      canonicalize{  max-iterations=10 max-num-rewrites=-1 region-simplify=normal test-convergence=false top-down=true},
      cse,
      gpu.module(
        canonicalize{max-iterations=10 max-num-rewrites=-1 region-simplify=normal test-convergence=false top-down=true}
      ),
      gpu.module(
        cse
      ),
      gpu.module(
        reconcile-unrealized-casts
      ),
      gpu-to-llvm{intersperse-sizes-for-kernels=false use-bare-pointers-for-host=false use-bare-pointers-for-kernels=false},
      """
        + inline
        + """,
      canonicalize{  max-iterations=10 max-num-rewrites=-1 region-simplify=normal test-convergence=false top-down=true},
      cse
    )
    """
    )


def _needs_llvm70_path(cc: str) -> bool:
    """Return True when libnvvm requires the LLVM 7 dialect of NVVM IR.

    The modern dialect (based on LLVM 20+) is used on Blackwell and later
    (sm_100+).  Everything below sm_100 requires the LLVM70 path which
    translates MLIR to the LLVM 7 dialect for the LLVM70 reader.
    """
    sm = int("".join(c for c in cc if c.isdigit()))
    return sm < 100


_llvm70_capi = None
# Keep AddDllDirectory handles alive while the CAPI DLL is loaded.
_llvm70_dll_dirs = []


@functools.cache
def _get_nvvm_ir_version():
    from numba_cuda_mlir.numba_cuda.cudadrv.nvvm import NVVM

    return NVVM().get_ir_version()


def _get_llvm70_capi():
    global _llvm70_capi
    if _llvm70_capi is not None:
        return _llvm70_capi

    from numba_cuda_mlir.tools import get_llvm70_capi_path

    capi_path = get_llvm70_capi_path()
    if not os.path.isfile(capi_path):
        raise FileNotFoundError(f"LLVM70 C API bridge not found at {capi_path}")
    if os.name == "nt":
        import numba_cuda_mlir._mlir._mlir_libs as _mlir_libs

        _llvm70_dll_dirs.extend(
            os.add_dll_directory(dll_dir)
            for dll_dir in {os.path.dirname(capi_path), _mlir_libs.__path__[0]}
            if os.path.isdir(dll_dir)
        )

    lib = ctypes.CDLL(capi_path)
    lib.llvm70_translate_gpu_module_from_op.restype = ctypes.c_int
    lib.llvm70_translate_gpu_module_from_op.argtypes = [
        ctypes.c_void_p,  # raw_op (Operation*)
        ctypes.c_char_p,  # chip
        ctypes.c_char_p,  # data_layout
        ctypes.c_char_p,  # libllvm
        ctypes.c_char_p,  # libnvvm
        ctypes.c_char_p,  # libdevice
        ctypes.c_int,  # gen_lto
        ctypes.c_int,  # gen_llvmir
        ctypes.c_int,  # opt_level
        ctypes.c_int,  # gen_lineinfo
        ctypes.c_int,  # nvvm_ir_major
        ctypes.c_int,  # nvvm_ir_minor
        ctypes.c_int,  # nvvm_dbg_major
        ctypes.c_int,  # nvvm_dbg_minor
        ctypes.POINTER(ctypes.c_void_p),  # out
        ctypes.POINTER(ctypes.c_size_t),  # out_len
        ctypes.POINTER(ctypes.c_void_p),  # nvvm_out
        ctypes.POINTER(ctypes.c_size_t),  # nvvm_out_len
        ctypes.POINTER(ctypes.c_char_p),  # err_out
    ]
    lib.llvm70_free.restype = None
    lib.llvm70_free.argtypes = [ctypes.c_void_p]
    _llvm70_capi = lib
    return lib


def _get_libnvvm_path() -> bytes:
    """Resolve libnvvm.so from the user's CTK (CUDA_HOME, conda, or pip)."""
    from numba_cuda_mlir.numba_cuda.cudadrv.libs import get_cudalib

    return get_cudalib("nvvm").encode()


def _get_op_ptr(op) -> ctypes.c_void_p:
    """Extract raw mlir::Operation* from a Python MLIR Operation via its capsule."""
    capsule = op._CAPIPtr
    ptr = ctypes.pythonapi.PyCapsule_GetPointer
    ptr.restype = ctypes.c_void_p
    ptr.argtypes = [ctypes.py_object, ctypes.c_char_p]
    return ptr(capsule, b"numba_cuda_mlir._mlir.ir.Operation._CAPIPtr")


def _get_single_gpu_module(module):
    from numba_cuda_mlir._mlir.dialects import gpu

    gpu_modules = [op for op in module.body if isinstance(op, gpu.GPUModuleOp)]
    if len(gpu_modules) != 1:
        raise ValueError(f"Expected exactly one gpu.module, found {len(gpu_modules)}")
    return gpu_modules[0]


def _call_llvm70_capi(module, target_options, gen_lto=False, gen_llvmir=False) -> bytes:
    """Compile MLIR gpu.module via in-process LLVM70 C API (raw Operation*)."""
    from numba_cuda_mlir.tools import get_gpu_compute_capability
    from numba_cuda_mlir.numba_cuda.cudadrv.libs import get_libdevice

    lib = _get_llvm70_capi()
    chip = target_options.get("chip", get_gpu_compute_capability())
    gpu_mod = _get_single_gpu_module(module)

    if target_options.get("dump_mlir") or target_options.get("dump"):
        print(f"=============== LLVM70 MLIR Module ===============\n\n{gpu_mod}\n")

    raw_op = _get_op_ptr(gpu_mod.operation)

    libllvm = os.environ.get("LIBLLVM7", "")
    if not libllvm:
        bundled_dir = os.path.join(os.path.dirname(__file__), "lib")
        if os.name == "nt":
            bundled_names = ("LLVM-C.dll", "LLVM.dll")
        else:
            bundled_names = ("libLLVM-7.so",)
        for bundled_name in bundled_names:
            bundled = os.path.join(bundled_dir, bundled_name)
            if os.path.isfile(bundled):
                libllvm = os.path.realpath(bundled)
                break

        # Fall back to a conda-forge libllvm7.1 package, whose shared library is
        # named libLLVM-7.1.so (its real soname). It lives in the environment's
        # lib dir, which is on our DSOs' RPATH, so dlopen resolves the soname.
        if not libllvm and os.name != "nt":
            libllvm = "libLLVM-7.1.so"

    if not libllvm:
        hint = "Set LIBLLVM7=/path/to/libLLVM-7.so"
        if os.name == "nt":
            hint = "Set LIBLLVM7=/path/to/LLVM-C.dll (or /path/to/LLVM.dll)"
        raise RuntimeError(f"LLVM70 path requires an LLVM 7 runtime library. {hint}")

    libnvvm = _get_libnvvm_path().decode()
    libdevice = get_libdevice()
    opt_level = int(target_options.get("opt_level", 2))
    if target_options.get("debug", False):
        debug_level = 2
    elif target_options.get("lineinfo", False):
        debug_level = 1
    else:
        debug_level = 0

    nvvm_ir_version = _get_nvvm_ir_version()
    out = ctypes.c_void_p()
    out_len = ctypes.c_size_t()
    nvvm_out = ctypes.c_void_p()
    nvvm_out_len = ctypes.c_size_t()
    err_out = ctypes.c_char_p()
    from numba_cuda_mlir.numba_cuda import config

    capture_nvvm = bool(config.CUDA_DUMP_NVVM) and not gen_llvmir

    rc = lib.llvm70_translate_gpu_module_from_op(
        raw_op,
        chip.encode(),
        None,
        libllvm.encode(),
        libnvvm.encode(),
        libdevice.encode(),
        1 if gen_lto else 0,
        1 if gen_llvmir else 0,
        opt_level,
        debug_level,
        nvvm_ir_version[0],
        nvvm_ir_version[1],
        nvvm_ir_version[2],
        nvvm_ir_version[3],
        ctypes.byref(out),
        ctypes.byref(out_len),
        ctypes.byref(nvvm_out) if capture_nvvm else None,
        ctypes.byref(nvvm_out_len) if capture_nvvm else None,
        ctypes.byref(err_out),
    )

    if rc != 0:
        msg = err_out.value.decode() if err_out.value else "unknown error"
        try:
            if nvvm_out.value:
                _maybe_dump_nvvm(ctypes.string_at(nvvm_out.value, nvvm_out_len.value))
        finally:
            if err_out.value:
                lib.llvm70_free(err_out)
            if nvvm_out.value:
                lib.llvm70_free(nvvm_out)
        raise RuntimeError(f"llvm70 translation failed: {msg}")

    try:
        result = ctypes.string_at(out.value, out_len.value)
        if nvvm_out.value:
            _maybe_dump_nvvm(ctypes.string_at(nvvm_out.value, nvvm_out_len.value))
        return result
    finally:
        if out.value:
            lib.llvm70_free(out)
        if nvvm_out.value:
            lib.llvm70_free(nvvm_out)


_nvvm_dump_counter = _LockedCounter()
_BITCODE_MAGIC = b"BC\xc0\xde"


def _maybe_dump_nvvm(nvvm_ir: bytes) -> None:
    """Dump the NVVM IR fed to libnvvm when configured."""
    from numba_cuda_mlir.numba_cuda import config

    target = config.CUDA_DUMP_NVVM
    if not target:
        return

    is_bitcode = nvvm_ir[:4] == _BITCODE_MAGIC
    if target.lower() in {"0", "false", "no", "off"}:
        return
    if target.lower() in {"1", "true", "yes", "stderr"}:
        if is_bitcode:
            print(
                f"=============== NVVM IR (bitcode, {len(nvvm_ir)} bytes) ===============\n"
                "Set NUMBA_CUDA_MLIR_DUMP_NVVM=<file|dir> to write the bitcode to disk.\n",
                file=sys.stderr,
            )
        else:
            print(
                f"=============== NVVM IR ===============\n\n"
                f"{nvvm_ir.decode('utf-8', errors='replace')}\n",
                file=sys.stderr,
            )
        return

    path = Path(target)
    if path.exists() and path.is_dir():
        extension = "bc" if is_bitcode else "ll"
        path = path / f"nvvm-{os.getpid()}-{next(_nvvm_dump_counter)}.{extension}"
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_bytes(nvvm_ir)


def _operation_to_text(operation, *, preserve_debug_info=False) -> str:
    if not preserve_debug_info:
        return str(operation)
    with StringIO() as sb:
        operation.print(enable_debug_info=True, file=sb)
        return sb.getvalue()


def _prepare_llvm_ir(module, dump=False, preserve_debug_info=False) -> bytes:
    """Translate gpu.module to LLVM IR and apply libnvvm compatibility downgrades."""
    from numba_cuda_mlir.tools import get_cuda_runtime_version

    gpu_mod = _get_single_gpu_module(module)
    gpu_mod.operation.attributes["llvm.data_layout"] = ir.StringAttr.get(NVPTX64_DATALAYOUT)
    gpu_mod.operation.attributes["llvm.target_triple"] = ir.StringAttr.get(NVPTX64_TRIPLE)
    ctk_major, ctk_minor = get_cuda_runtime_version()
    nvvm_ir_version = _get_nvvm_ir_version()

    if os.name == "nt":
        nvvm_ir = translate_gpu_module_to_libnvvm_ir(
            _operation_to_text(gpu_mod.operation, preserve_debug_info=preserve_debug_info),
            ctk_major,
            ctk_minor,
            nvvm_ir_version,
            dump=dump,
            emit_text_ir=preserve_debug_info,
        )
        _maybe_dump_nvvm(nvvm_ir)
        return nvvm_ir

    from numba_cuda_mlir._cext import downgrade_for_libnvvm

    llvm_mod, llvm_ctx = translate_to_llvmir(gpu_mod.operation)

    if dump:
        print(f"=============== LLVM IR ===============\n\n{dump_llvmir(llvm_mod)}\n\n")

    nvvm_ir = downgrade_for_libnvvm(
        llvm_mod,
        llvm_ctx,
        ctk_major,
        ctk_minor,
        nvvm_ir_version[0],
        nvvm_ir_version[1],
        nvvm_ir_version[2],
        nvvm_ir_version[3],
        LLVM_C_LIB_PATH,
    )
    _maybe_dump_nvvm(nvvm_ir)
    return nvvm_ir


def _nvvm_options(cc: str, target_options=None, **extra) -> dict:
    """Build libnvvm CompilationUnit options from arch + target options."""
    opts = {"arch": f"compute_{cc}", **extra}
    if target_options is None:
        return opts
    if target_options.get("fastmath"):
        opts.update({"ftz": True, "fma": True, "prec_div": False, "prec_sqrt": False})
    # Note: we intentionally omit -g and -generate-line-info here.
    # Our MLIR pipeline embeds DWARF metadata (DICompileUnit, DISubprogram, DILocation)
    # into the LLVM IR when debug=True or lineinfo=True. libnvvm honors that metadata
    # to produce debug sections and .loc/.file PTX directives automatically. Passing
    # either flag to libnvvm conflicts with pre-existing debug metadata, causing
    # NVVM_ERROR_IR_VERSION_MISMATCH.
    opt = target_options.get("opt")
    if opt is False or opt == 0:
        opts["opt"] = 0
    return opts


def _compile_to_ptx(llvm_ir: bytes, cc: str, libdevice, nvvm_opts=None) -> bytes:
    """Compile LLVM IR to PTX via libnvvm."""
    if nvvm_opts is None:
        nvvm_opts = {"arch": f"compute_{cc}"}
    cu = CompilationUnit(nvvm_opts)
    cu.add_module(llvm_ir)
    cu.verify()
    cu.lazy_add_module(libdevice.get())
    return cu.compile()


def _compile_to_ltoir(llvm_ir: bytes, libdevice, nvvm_opts: dict) -> bytes:
    cu = CompilationUnit({**nvvm_opts, "gen-lto": None})
    cu.add_module(llvm_ir)
    cu.verify()
    cu.lazy_add_module(libdevice.get())
    return cu.compile()


def _get_ltoir(cres, target_options) -> bytes:
    ltoir = cres.metadata.get("ltoir")
    if ltoir:
        return ltoir

    with context.get_context():
        module = ir.Module.parse(cres.metadata["mlir_module_optimized"])
        run_pre_codegen_patterns(module)

        chip = target_options.get("chip")
        if not chip:
            from numba_cuda_mlir.tools import get_gpu_compute_capability

            chip = get_gpu_compute_capability()
        cc = chip.replace("sm_", "")

        if _needs_llvm70_path(cc):
            ltoir = _call_llvm70_capi(module, target_options, gen_lto=True)
        else:
            llvm_ir = _prepare_llvm_ir(
                module,
                dump=target_options.get("dump_llvmir", False),
                preserve_debug_info=target_options.get("debug", False)
                or target_options.get("lineinfo", False),
            )
            from numba_cuda_mlir.numba_cuda.cudadrv.nvvm import LibDevice

            libdevice = LibDevice()
            nvvm_opts = _nvvm_options(cc, target_options)
            ltoir = _compile_to_ltoir(llvm_ir, libdevice, nvvm_opts)

    cres.metadata["ltoir"] = ltoir
    return ltoir


def get_llvmir(cres, target_options=None) -> str:
    """Return architecture-natural LLVM IR lazily for inspection."""
    if target_options is None:
        target_options = cres.metadata["targetoptions"]

    with context.get_context():
        module = ir.Module.parse(cres.metadata["mlir_module_optimized"])
        run_pre_codegen_patterns(module)

        chip = target_options.get("chip")
        if not chip:
            from numba_cuda_mlir.tools import get_gpu_compute_capability

            chip = get_gpu_compute_capability()
        cc = chip.replace("sm_", "")

        if _needs_llvm70_path(cc):
            return _call_llvm70_capi(
                module,
                target_options,
                gen_lto=target_options.get("lto", False),
                gen_llvmir=True,
            ).decode("utf-8")

        gpu_mod = _get_single_gpu_module(module)
        gpu_mod.operation.attributes["llvm.data_layout"] = ir.StringAttr.get(NVPTX64_DATALAYOUT)
        gpu_mod.operation.attributes["llvm.target_triple"] = ir.StringAttr.get(NVPTX64_TRIPLE)

        if os.name == "nt":
            from numba_cuda_mlir.tools import get_cuda_runtime_version

            ctk_major, ctk_minor = get_cuda_runtime_version()
            llvm_ir = translate_gpu_module_to_libnvvm_ir(
                _operation_to_text(
                    gpu_mod.operation,
                    preserve_debug_info=target_options.get("debug", False)
                    or target_options.get("lineinfo", False),
                ),
                ctk_major,
                ctk_minor,
                _get_nvvm_ir_version(),
                inspect_llvmir=True,
            )
            return llvm_ir.decode("utf-8")

        return translate_to_llvmir_text(gpu_mod.operation)


def get_ptx(cres, target_options=None) -> str:
    """Return regular PTX lazily for inspection without doing it during LTO compilation."""
    ptx = cres.metadata.get("ptx")
    if ptx:
        return ptx
    if target_options is None:
        target_options = cres.metadata["targetoptions"]

    with context.get_context():
        module = ir.Module.parse(cres.metadata["mlir_module_optimized"])
        run_pre_codegen_patterns(module)

        chip = target_options.get("chip")
        if not chip:
            from numba_cuda_mlir.tools import get_gpu_compute_capability

            chip = get_gpu_compute_capability()
        cc = chip.replace("sm_", "")

        if _needs_llvm70_path(cc):
            ptx = _call_llvm70_capi(module, target_options)
        else:
            llvm_ir = _prepare_llvm_ir(
                module,
                dump=target_options.get("dump_llvmir", False),
                preserve_debug_info=target_options.get("debug", False)
                or target_options.get("lineinfo", False),
            )
            from numba_cuda_mlir.numba_cuda.cudadrv.nvvm import LibDevice

            libdevice = LibDevice()
            nvvm_opts = _nvvm_options(cc, target_options)
            ptx = _compile_to_ptx(llvm_ir, cc, libdevice, nvvm_opts)

    return ptx.decode()


def _dump_module(mod, header):
    print(header, end="\n\n")
    # Include loc and #llvm.di_* when present (e.g. debug/lineinfo).
    mod.operation.print(enable_debug_info=True)
    print("\n\n")


def _find_dbg_var_name(loc):
    """Extract dbg variable name from nested locations."""
    if isinstance(loc, ir.NameLoc):
        if loc.name_str.startswith("dbg_var:"):
            return loc.name_str[len("dbg_var:") :]
        return _find_dbg_var_name(loc.child_loc)
    if isinstance(loc, ir.FusedLoc):
        for nested_loc in loc.locations:
            name = _find_dbg_var_name(nested_loc)
            if name is not None:
                return name
    return None


def _strip_dbg_var_nameloc(loc):
    """Strip dbg_var: NameLoc entries from a location tree."""
    if isinstance(loc, ir.NameLoc):
        if loc.name_str.startswith("dbg_var:"):
            return None
        child_loc = _strip_dbg_var_nameloc(loc.child_loc)
        if child_loc is None:
            return ir.Location.name(loc.name_str)
        return ir.Location.name(loc.name_str, child_loc)
    if not isinstance(loc, ir.FusedLoc):
        return loc
    stripped = []
    for nested_loc in loc.locations:
        new_loc = _strip_dbg_var_nameloc(nested_loc)
        if new_loc is not None:
            stripped.append(new_loc)
    if not stripped:
        return loc
    if len(stripped) == 1:
        return stripped[0]
    return ir.Location.fused(stripped)


def _cleanup_deferred_dbg_attrs(module_attrs):
    """Remove internal deferred-debug module attributes."""
    expr_key = "metadata.dbg_declare_expr"
    var_key_prefix = "metadata.dbg_declare_var_"
    for key in [k for k in module_attrs.keys() if k.startswith(var_key_prefix)]:
        del module_attrs[key]
    if expr_key in module_attrs:
        del module_attrs[expr_key]


def _emit_deferred_dbg_declares(module):
    """A deferred emission of dbg.declare for variables tagged during lowering.

    Lowering tags memref.alloca ops with ``dbg_var:<name>`` NameLoc for variables
    that need to be emitted as dbg.declare after the base pipeline. This helper
    emits llvm.intr.dbg.declare for tagged allocas in module-local DI scope.
    """
    module_attrs = module.operation.attributes
    var_key_prefix = "metadata.dbg_declare_var_"
    expr_key = "metadata.dbg_declare_expr"
    expr_attr = module_attrs[expr_key] if expr_key in module_attrs else None
    if expr_attr is None:
        _cleanup_deferred_dbg_attrs(module_attrs)
        return False

    tagged_vars = []
    emitted_vars = set()
    for op in find_ops(module, lambda o: o.name == "llvm.alloca"):
        loc = op.location
        var_name = _find_dbg_var_name(loc)
        if var_name is None:
            continue
        if var_name in emitted_vars:
            continue
        key = f"{var_key_prefix}{var_name}"
        var_attr = module_attrs[key] if key in module_attrs else None
        if var_attr is None:
            continue
        tagged_vars.append((op, var_name, var_attr))
        emitted_vars.add(var_name)

    for op, var_name, var_attr in tagged_vars:
        # Recover the clean/original location (without dbg_var tags) and use it
        # consistently for both the alloca and dbg.declare ops.
        op.location = _strip_dbg_var_nameloc(op.location)
        with ir.InsertionPoint.after(op), op.location:
            llvm.intr_dbg_declare(op.result, var_attr, location_expr=expr_attr)
        trace("Deferred dbg.declare emitted for %s", var_name)
    _cleanup_deferred_dbg_attrs(module_attrs)
    return bool(tagged_vars)


def get_lto_ptx(cres, linker=None, target_options=None) -> str:
    """Return PTX after LTO without requiring it during normal compilation."""
    from numba_cuda_mlir.linker import Linker, _link_item_is_cuda_source

    ptx = cres.metadata.get("lto_ptx")
    if ptx:
        return ptx
    if target_options is None:
        target_options = cres.metadata["targetoptions"]
    if linker is None:
        linker = cres.metadata.get("linker")
    if linker is None:
        from numba_cuda_mlir.tools import get_gpu_compute_capability, parse_compute_capability

        chip = target_options.get("chip")
        if chip:
            cc = parse_compute_capability(chip)
            arch = chip
        else:
            cc = get_gpu_compute_capability(tuple)
            arch = get_gpu_compute_capability(str)

        linker = Linker(
            cc=cc,
            arch=arch,
            verbose=target_options.get("dump", False),
            debug=target_options.get("debug", False),
            lineinfo=target_options.get("lineinfo", False),
            lto=True,
            ftz=target_options.get("fastmath") or None,
            prec_div=False if target_options.get("fastmath") else None,
            prec_sqrt=False if target_options.get("fastmath") else None,
            fma=target_options.get("fastmath") or None,
            optimization_level=int(target_options.get("opt_level", 3)),
            ptxas_options=target_options.get("ptxas_options", None),
            max_registers=target_options.get("max_registers", None),
        )

    diag_linker = linker.recreate_with_lto(lto=True, ltoir_only=True)
    diag_linker.additional_flags = ["-ptx"]
    diag_linker.add_ltoir(_get_ltoir(cres, target_options))
    link_plan = cres.metadata.get("link_plan")
    for link_file in cres.metadata.get(
        "linked_external_link_items", target_options.get("link", [])
    ):
        if (
            link_plan is not None
            and link_plan.compile_new_inputs_as_ltoir
            and _link_item_is_cuda_source(link_file)
        ):
            continue
        diag_linker.add_file_guess_ext(link_file, ignore_nonlto=True)
    if cres.metadata.get("needs_nrt") and not cres.metadata.get("nrt_inline"):
        _maybe_link_nrt(diag_linker)
    return diag_linker.get_linked_ptx().decode("utf-8")


def _dump_lto_assembly(cres, linker, target_options):
    """Diagnostic LTO-to-PTX link to dump post-LTO assembly and warn about
    non-LTO linkables, mirroring CUDACodeLibrary.get_cubin() / get_lto_ptx()."""
    ptx_after_lto = get_lto_ptx(cres, linker, target_options)
    name = cres.fndesc.qualname
    print(("ASSEMBLY (AFTER LTO) %s" % name).center(80, "-"))
    print(ptx_after_lto)
    print("=" * 80)


def optimize(cres):
    with context.get_context():
        target_options = cres.metadata["targetoptions"]
        dump_mlir = target_options.get("dump_mlir", False) or target_options.get("dump", False)
        module = cres.metadata["mlir_module"]
        # Materialize the pre-optimization string before passes mutate the module,
        # so that inspect_mlir() and other consumers can still access it.
        from numba_cuda_mlir.mlir_lowering import get_mlir_module_str

        get_mlir_module_str(cres.metadata)

        if dump_mlir:
            _dump_module(module, "=============== MLIR Module ===============")

        pm = PassManager.parse(get_base_pipeline())
        pm.enable_ir_printing(
            print_before_all=target_options.get("print_before_all", False),
            print_after_all=target_options.get("print_after_all", False),
        )
        pm.run(module.operation)

        if target_options.get("debug"):
            _emit_deferred_dbg_declares(module)
        if target_options.get("debug") or target_options.get("lineinfo"):
            with StringIO() as sb:
                module.operation.print(enable_debug_info=True, file=sb)
                cres.metadata["mlir_module_optimized"] = sb.getvalue()
        else:
            cres.metadata["mlir_module_optimized"] = str(module)
        if dump_mlir:
            _dump_module(module, "=============== Optimized MLIR Module ===============")

        run_pre_codegen_patterns(module)
        if dump_mlir:
            _dump_module(
                module,
                "=============== Optimized MLIR Module (after pre-codegen patterns) ===============",
            )

        chip = target_options.get("chip")
        if not chip:
            from numba_cuda_mlir.tools import get_gpu_compute_capability

            chip = get_gpu_compute_capability()
        cc = chip.replace("sm_", "")
        link_plan = cres.metadata.get("link_plan")
        is_lto = (
            link_plan.compile_new_inputs_as_ltoir
            if link_plan is not None
            else target_options.get("lto", False)
            or target_options.get("_compile_output") == "ltoir"
        )

        from numba_cuda_mlir.numba_cuda.cudadrv.nvvm import LibDevice

        use_llvm70 = _needs_llvm70_path(cc)

        libdevice = LibDevice()
        nvvm_opts = _nvvm_options(cc, target_options)
        if use_llvm70:
            llvm_ir = None
        else:
            llvm_ir = _prepare_llvm_ir(
                module,
                dump=target_options.get("dump_llvmir", False),
                preserve_debug_info=target_options.get("debug", False)
                or target_options.get("lineinfo", False),
            )

        base_linker = cres.metadata["linker"]

        if is_lto:
            if use_llvm70:
                ltoir = _call_llvm70_capi(module, target_options, gen_lto=True)
            else:
                ltoir = _compile_to_ltoir(llvm_ir, libdevice, nvvm_opts)
            cres.metadata["ltoir"] = ltoir
            cres.metadata["ptx"] = ""
            # The kernel's own LTO-IR must be first in the link chain, ahead of
            # any device-function and library LTO-IRs accumulated during
            # lowering, so nvJitLink resolves symbols against the kernel first.
            linker = base_linker.recreate_with_lto(kernel_ltoir=ltoir)
            if target_options.get("dump_ptx", False):
                cres.metadata["lto_ptx"] = get_lto_ptx(cres, linker, target_options)
                print(f"=============== PTX ===============\n\n{cres.metadata['lto_ptx']}\n\n")
        else:
            linker = base_linker.recreate_with_lto()
            if use_llvm70:
                ptx = _call_llvm70_capi(module, target_options)
            else:
                ptx = _compile_to_ptx(llvm_ir, cc, libdevice, nvvm_opts)
            cres.metadata["ptx"] = ptx.decode()
            if target_options.get("dump_ptx", False):
                print(f"=============== PTX ===============\n\n{cres.metadata['ptx']}\n\n")
            linker.add_ptx(ptx)

        code = linker.complete()
        cres.metadata["cubin"] = code.code

        if is_lto:
            from numba_cuda_mlir.numba_cuda import config

            if config.DUMP_ASSEMBLY:
                _dump_lto_assembly(cres, linker, target_options)

        if target_options.get("dump_cubin", False):
            print(f"=============== Cubin ===============\n\n{code.code}\n\n")

        if cres.metadata.get("needs_nrt"):
            cres.metadata.setdefault("setup_callbacks", [])
            if _nrt_memsys_setup_callback not in cres.metadata["setup_callbacks"]:
                cres.metadata["setup_callbacks"].append(_nrt_memsys_setup_callback)

        # TODO: parse CC from the object and ensure it's not greater than the
        # greatest supported CC via _get_gpu_compute_capability()
        cres.metadata["func_name"] = generate_mangled_name(
            cres.fndesc.qualname, cres.fndesc.argtypes
        )

        cres.library._ptx = cres.metadata["ptx"]
        cres.library._mlir_str = cres.metadata["mlir_module_optimized"]



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
import numba_cuda_mlir.cuda.libdevice as libdevice
import numba_cuda_mlir.cuda.libdevicefuncs as libdevicefuncs
from numba_cuda_mlir.numba_cuda.typing.templates import (
    AttributeTemplate,
    ConcreteTemplate,
    Registry,
)
from numba_cuda_mlir.numba_cuda import types

registry = Registry()
_mapping = dict[str, ConcreteTemplate]()


def libdevice_declare(pyfunc, py_sig):
    class Libdevice_function(ConcreteTemplate):
        key = pyfunc
        cases = [py_sig]

    registry.register_global(pyfunc)(Libdevice_function)
    return Libdevice_function


def _libdevice_register():
    """
    Register type declarations and lowerings for all libdevice functions.
    """
    for descriptor in libdevicefuncs.libdevice_descriptors():
        py_sig = descriptor.py_sig
        name = descriptor.py_name
        pyfunc = getattr(libdevice, descriptor.py_name)
        Libdevice_function = libdevice_declare(pyfunc, py_sig)
        _mapping[name] = Libdevice_function


@registry.register_attr
class LibdeviceModuleTemplate(AttributeTemplate):
    import numba_cuda_mlir.cuda.libdevice as libdevice

    key = types.Module(libdevice)

    def resolve(self, mod, attr):
        if template := _mapping.get(attr):
            return types.Function(template)
        raise AttributeError(f"Libdevice module has no attribute {attr!r}")


_libdevice_register()



#  Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
#  See https://llvm.org/LICENSE.txt for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
import operator
from itertools import accumulate
from typing import Optional

from ._memref_ops_gen import *
from ._memref_ops_gen import _Dialect
from ._ods_common import _dispatch_mixed_values, MixedValues
from ..ir import (
    IndexType,
    IntegerType,
    MemRefType,
    ShapedType,
    StridedLayoutAttr,
    Value,
)
from . import arith


def _is_constant_int_like(i):
    return (
        isinstance(i, Value)
        and isinstance(i.owner, arith.ConstantOp)
        and isinstance(i.type, (IntegerType, IndexType))
    )


def _is_static_int_like(i):
    return (
        isinstance(i, int) and not ShapedType.is_dynamic_size(i)
    ) or _is_constant_int_like(i)


def _infer_memref_subview_result_type(
    source_memref_type, offsets, static_sizes, static_strides
):
    source_strides, source_offset = source_memref_type.get_strides_and_offset()
    # "canonicalize" from tuple|list -> list
    offsets, static_sizes, static_strides, source_strides = map(
        list, (offsets, static_sizes, static_strides, source_strides)
    )

    if not all(
        all(_is_static_int_like(i) for i in s)
        for s in [
            static_sizes,
            static_strides,
            source_strides,
        ]
    ):
        raise ValueError(
            "Only inferring from python or mlir integer constant is supported."
        )

    for s in [offsets, static_sizes, static_strides]:
        for idx, i in enumerate(s):
            if _is_constant_int_like(i):
                s[idx] = i.owner.opview.literal_value

    if any(not _is_static_int_like(i) for i in offsets + [source_offset]):
        target_offset = ShapedType.get_dynamic_size()
    else:
        target_offset = source_offset
        for offset, target_stride in zip(offsets, source_strides):
            target_offset += offset * target_stride

    target_strides = []
    for source_stride, static_stride in zip(source_strides, static_strides):
        target_strides.append(source_stride * static_stride)

    # If default striding then no need to complicate things for downstream ops (e.g., expand_shape).
    default_strides = list(accumulate(static_sizes[1:][::-1], operator.mul))[::-1] + [1]
    if target_strides == default_strides and target_offset == 0:
        layout = None
    else:
        layout = StridedLayoutAttr.get(target_offset, target_strides)
    return (
        offsets,
        static_sizes,
        static_strides,
        MemRefType.get(
            static_sizes,
            source_memref_type.element_type,
            layout,
            source_memref_type.memory_space,
        ),
    )


_generated_subview = subview


def subview(
    source: Value,
    offsets: MixedValues,
    sizes: MixedValues,
    strides: MixedValues,
    *,
    result_type: Optional[MemRefType] = None,
    loc=None,
    ip=None,
):
    if offsets is None:
        offsets = []
    if sizes is None:
        sizes = []
    if strides is None:
        strides = []
    source_strides, source_offset = source.type.get_strides_and_offset()
    if result_type is None and all(
        all(_is_static_int_like(i) for i in s) for s in [sizes, strides, source_strides]
    ):
        # If any are arith.constant results then this will canonicalize to python int
        # (which can then be used to fully specify the subview).
        (
            offsets,
            sizes,
            strides,
            result_type,
        ) = _infer_memref_subview_result_type(source.type, offsets, sizes, strides)
    elif result_type is None:
        raise ValueError(
            "mixed static/dynamic offset/sizes/strides requires explicit result type."
        )

    offsets, _packed_offsets, static_offsets = _dispatch_mixed_values(offsets)
    sizes, _packed_sizes, static_sizes = _dispatch_mixed_values(sizes)
    strides, _packed_strides, static_strides = _dispatch_mixed_values(strides)

    return _generated_subview(
        result_type,
        source,
        offsets,
        sizes,
        strides,
        static_offsets,
        static_sizes,
        static_strides,
        loc=loc,
        ip=ip,
    )



#  Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
#  See https://llvm.org/LICENSE.txt for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception

from .._bufferization_transform_ops_gen import *
from .._bufferization_transform_ops_gen import _Dialect

try:
    from ...ir import *
    from ...dialects import transform
    from .._ods_common import _cext as _ods_cext
except ImportError as e:
    raise RuntimeError("Error loading imports from extension module") from e

from enum import Enum
from typing import Optional, overload, Union


@_ods_cext.register_operation(_Dialect, replace=True)
class EmptyTensorToAllocTensorOp(EmptyTensorToAllocTensorOp):
    """Specialization for EmptyTensorToAllocTensorOp class."""

    @overload
    def __init__(
        self,
        transformed_type: Type,
        target: Union[Operation, OpView, Value],
        *,
        loc=None,
        ip=None,
    ):
        ...

    @overload
    def __init__(self, target: Union[Operation, OpView, Value], *, loc=None, ip=None):
        ...

    def __init__(
        self,
        transformed_type_or_target: Type,
        target_or_none: Optional[Union[Operation, OpView, Value]] = None,
        *,
        loc=None,
        ip=None,
    ):
        if isinstance(transformed_type_or_target, Type):
            transformed_type = transformed_type_or_target
            target = target_or_none
        else:
            transformed_type = transform.OperationType.get("bufferization.alloc_tensor")
            target = transformed_type_or_target

        super().__init__(
            transformed_type,
            target,
            loc=loc,
            ip=ip,
        )


@_ods_cext.register_operation(_Dialect, replace=True)
class OneShotBufferizeOp(OneShotBufferizeOp):
    """Specialization for OneShotBufferizeOp class."""

    @overload
    def __init__(
        self,
        transformed_type: Type,
        target: Union[Operation, OpView, Value],
        *,
        allow_return_allocs_from_loops: Optional[bool] = None,
        allow_unknown_ops: Optional[bool] = None,
        bufferize_function_boundaries: Optional[bool] = None,
        function_boundary_type_conversion: Optional[Enum] = None,
        memcpy_op: Optional[str] = None,
        print_conflicts: Optional[bool] = None,
        test_analysis_only: Optional[bool] = None,
        loc=None,
        ip=None,
    ):
        ...

    @overload
    def __init__(
        self,
        target: Union[Operation, OpView, Value],
        *,
        allow_return_allocs_from_loops: Optional[bool] = None,
        allow_unknown_ops: Optional[bool] = None,
        bufferize_function_boundaries: Optional[bool] = None,
        function_boundary_type_conversion: Optional[Enum] = None,
        memcpy_op: Optional[str] = None,
        print_conflicts: Optional[bool] = None,
        test_analysis_only: Optional[bool] = None,
        loc=None,
        ip=None,
    ):
        ...

    def __init__(
        self,
        transformed_type_or_target: Type,
        target_or_none: Optional[Union[Operation, OpView, Value]] = None,
        *,
        allow_return_allocs_from_loops: Optional[bool] = None,
        allow_unknown_ops: Optional[bool] = None,
        bufferize_function_boundaries: Optional[bool] = None,
        function_boundary_type_conversion: Optional[Enum] = None,
        memcpy_op: Optional[str] = None,
        print_conflicts: Optional[bool] = None,
        test_analysis_only: Optional[bool] = None,
        loc=None,
        ip=None,
    ):
        if isinstance(transformed_type_or_target, Type):
            transformed_type = transformed_type_or_target
            target = target_or_none
        else:
            transformed_type = transform.AnyOpType.get()
            target = transformed_type_or_target

        super().__init__(
            transformed_type,
            target,
            allow_return_allocs_from_loops=allow_return_allocs_from_loops,
            allow_unknown_ops=allow_unknown_ops,
            bufferize_function_boundaries=bufferize_function_boundaries,
            function_boundary_type_conversion=function_boundary_type_conversion,
            memcpy_op=memcpy_op,
            print_conflicts=print_conflicts,
            test_analysis_only=test_analysis_only,
            loc=loc,
            ip=ip,
        )




# Autogenerated by mlir-tblgen; don't manually edit.

from enum import IntEnum, auto, IntFlag
from ._ods_common import _cext as _ods_cext
from ..ir import register_attribute_builder
_ods_ir = _ods_cext.ir

class AddressSpace(IntEnum):
    """AMDGPU-specific address spaces"""

    FatRawBuffer = 0
    BufferRsrc = 1
    FatStructuredBuffer = 2

    def __str__(self):
        if self is AddressSpace.FatRawBuffer:
            return "fat_raw_buffer"
        if self is AddressSpace.BufferRsrc:
            return "buffer_rsrc"
        if self is AddressSpace.FatStructuredBuffer:
            return "fat_structured_buffer"
        raise ValueError("Unknown AddressSpace enum entry.")



class Scope(IntEnum):
    """AMDGPU-specific cache scopes. WGP - workgroup processor (CUs); SE - shader engine (GL2); DEV - device; SYS - system"""

    WGP = 0
    SE = 1
    DEV = 2
    SYS = 3

    def __str__(self):
        if self is Scope.WGP:
            return "WGP"
        if self is Scope.SE:
            return "SE"
        if self is Scope.DEV:
            return "DEV"
        if self is Scope.SYS:
            return "SYS"
        raise ValueError("Unknown Scope enum entry.")



class DPPPerm(IntEnum):
    """The possible permutations for a DPP operation"""

    quad_perm = 0
    row_shl = 1
    row_shr = 2
    row_ror = 3
    wave_shl = 4
    wave_shr = 5
    wave_ror = 6
    wave_rol = 7
    row_mirror = 8
    row_half_mirror = 9
    row_bcast_15 = 10
    row_bcast_31 = 11

    def __str__(self):
        if self is DPPPerm.quad_perm:
            return "quad_perm"
        if self is DPPPerm.row_shl:
            return "row_shl"
        if self is DPPPerm.row_shr:
            return "row_shr"
        if self is DPPPerm.row_ror:
            return "row_ror"
        if self is DPPPerm.wave_shl:
            return "wave_shl"
        if self is DPPPerm.wave_shr:
            return "wave_shr"
        if self is DPPPerm.wave_ror:
            return "wave_ror"
        if self is DPPPerm.wave_rol:
            return "wave_rol"
        if self is DPPPerm.row_mirror:
            return "row_mirror"
        if self is DPPPerm.row_half_mirror:
            return "row_half_mirror"
        if self is DPPPerm.row_bcast_15:
            return "row_bcast_15"
        if self is DPPPerm.row_bcast_31:
            return "row_bcast_31"
        raise ValueError("Unknown DPPPerm enum entry.")



class LoadTemporalHint(IntEnum):
    """AMDGPU-specific prefetch temporal hints for load instructions. RT - regular temporal for both near and far caches; NT - non-temporal for both near and far caches; HT - high-priority temporal for both near and far caches; LU - last-use; NT_RT - non-temporal for near cache(s) and regular for far caches; RT_NT - regular for near cache(s) and non-temporal for far caches; NT_HT - non-temporal for near cache(s) and high-priority temporal for far caches"""

    RT = 0
    NT = 1
    HT = 2
    LU = 3
    NT_RT = 4
    RT_NT = 5
    NT_HT = 6

    def __str__(self):
        if self is LoadTemporalHint.RT:
            return "RT"
        if self is LoadTemporalHint.NT:
            return "NT"
        if self is LoadTemporalHint.HT:
            return "HT"
        if self is LoadTemporalHint.LU:
            return "LU"
        if self is LoadTemporalHint.NT_RT:
            return "NT_RT"
        if self is LoadTemporalHint.RT_NT:
            return "RT_NT"
        if self is LoadTemporalHint.NT_HT:
            return "NT_HT"
        raise ValueError("Unknown LoadTemporalHint enum entry.")



class MFMAPermB(IntEnum):
    """The possible permutations of the lanes storing B available in an MFMA"""

    none = 0
    bcast_first_32 = 1
    bcast_second_32 = 2
    rotate_16_right = 3
    bcast_first_16 = 4
    bcast_second_16 = 5
    bcast_third_16 = 6
    bcast_fourth_16 = 7

    def __str__(self):
        if self is MFMAPermB.none:
            return "none"
        if self is MFMAPermB.bcast_first_32:
            return "bcast_first_32"
        if self is MFMAPermB.bcast_second_32:
            return "bcast_second_32"
        if self is MFMAPermB.rotate_16_right:
            return "rotate_16_right"
        if self is MFMAPermB.bcast_first_16:
            return "bcast_first_16"
        if self is MFMAPermB.bcast_second_16:
            return "bcast_second_16"
        if self is MFMAPermB.bcast_third_16:
            return "bcast_third_16"
        if self is MFMAPermB.bcast_fourth_16:
            return "bcast_fourth_16"
        raise ValueError("Unknown MFMAPermB enum entry.")



class sched_barrier_opt_enum(IntFlag):
    """The possible options for scheduling barriers"""

    none = 0
    non_mem_non_sideffect = 1
    valu = 2
    salu = 4
    mfma_wmma = 8
    all_vmem = 16
    vmem_read = 32
    vmem_write = 64
    all_ds = 128
    ds_read = 256
    ds_write = 512
    transcendental = 1024

    def __iter__(self):
        return iter([case for case in type(self) if (self & case) is case and self is not case])
    def __len__(self):
        return bin(self).count("1")

    def __str__(self):
        if len(self) > 1:
            return "|".join(map(str, self))
        if self is sched_barrier_opt_enum.none:
            return "none"
        if self is sched_barrier_opt_enum.non_mem_non_sideffect:
            return "non_mem_non_sideffect"
        if self is sched_barrier_opt_enum.valu:
            return "valu"
        if self is sched_barrier_opt_enum.salu:
            return "salu"
        if self is sched_barrier_opt_enum.mfma_wmma:
            return "mfma_wmma"
        if self is sched_barrier_opt_enum.all_vmem:
            return "all_vmem"
        if self is sched_barrier_opt_enum.vmem_read:
            return "vmem_read"
        if self is sched_barrier_opt_enum.vmem_write:
            return "vmem_write"
        if self is sched_barrier_opt_enum.all_ds:
            return "all_ds"
        if self is sched_barrier_opt_enum.ds_read:
            return "ds_read"
        if self is sched_barrier_opt_enum.ds_write:
            return "ds_write"
        if self is sched_barrier_opt_enum.transcendental:
            return "transcendental"
        raise ValueError("Unknown sched_barrier_opt_enum enum entry.")



@register_attribute_builder("amdgpu.AMDGPU_AddressSpaceAttr")
def _amdgpu_addressspaceattr(x, context):
    return _ods_ir.Attribute.parse(f'#amdgpu.address_space<{str(x)}>', context=context)

@register_attribute_builder("amdgpu.AMDGPU_CacheScopeAttr")
def _amdgpu_cachescopeattr(x, context):
    return _ods_ir.Attribute.parse(f'#amdgpu<cache_scope {str(x)}>', context=context)

@register_attribute_builder("amdgpu.AMDGPU_DPPPermAttr")
def _amdgpu_dpppermattr(x, context):
    return _ods_ir.Attribute.parse(f'#amdgpu<dpp_perm {str(x)}>', context=context)

@register_attribute_builder("amdgpu.AMDGPU_LoadTemporalHintAttr")
def _amdgpu_loadtemporalhintattr(x, context):
    return _ods_ir.Attribute.parse(f'#amdgpu<load_temporal_hint {str(x)}>', context=context)

@register_attribute_builder("amdgpu.AMDGPU_MFMAPermBAttr")
def _amdgpu_mfmapermbattr(x, context):
    return _ods_ir.Attribute.parse(f'#amdgpu<mfma_perm_b {str(x)}>', context=context)

@register_attribute_builder("amdgpu.AMDGPU_SchedBarrierOpOptAttr")
def _amdgpu_schedbarrieropoptattr(x, context):
    return _ods_ir.Attribute.parse(f'#amdgpu.sched_barrier_opt<{str(x)}>', context=context)




# SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import abc
from contextlib import contextmanager
from collections import defaultdict, namedtuple
from copy import copy
import warnings

from numba_cuda_mlir.numba_cuda.core import typeinfer
from numba_cuda_mlir.numba_cuda.core import (
    errors,
)
from numba_cuda_mlir.numba_cuda.core import ir
from numba_cuda_mlir.numba_cuda import typing, types
from numba_cuda_mlir.numba_cuda.core.compiler_machinery import (
    FunctionPass,
    LoweringPass,
    AnalysisPass,
    register_pass,
)
from numba_cuda_mlir.numba_cuda.core.annotations import type_annotations
from numba_cuda_mlir.numba_cuda.core.ir_utils import (
    raise_on_unsupported_feature,
    warn_deprecated,
    check_and_legalize_ir,
    guard,
    dead_code_elimination,
    simplify_CFG,
    get_definition,
    compute_cfg_from_blocks,
    is_operator_or_getitem,
)
from numba_cuda_mlir.numba_cuda.core.callconv import CUDACallConv

from numba_cuda_mlir.numba_cuda.core import postproc, rewrites, funcdesc, config
from numba_cuda_mlir.numba_cuda.core.options import InlineOptions


# Outputs of type inference pass
_TypingResults = namedtuple(
    "_TypingResults",
    [
        "typemap",
        "return_type",
        "calltypes",
        "typing_errors",
    ],
)


@contextmanager
def fallback_context(state, msg):
    """
    Wraps code that would signal a fallback to object mode
    """
    try:
        yield
    except Exception as e:
        if not state.status.can_fallback:
            raise
        else:
            # Clear all references attached to the traceback
            e = e.with_traceback(None)
            # this emits a warning containing the error message body in the
            # case of fallback from npm to objmode
            loop_lift = "" if state.flags.enable_looplift else "OUT"
            warnings.warn_explicit(
                "Compilation is falling back to object mode "
                f"WITH{loop_lift} looplifting enabled because {msg} due to: {e}",
                errors.NumbaWarning,
                state.func_id.filename,
                state.func_id.firstlineno,
            )
            raise


def type_inference_stage(
    typingctx,
    targetctx,
    interp,
    args,
    return_type,
    locals=None,
    raise_errors=True,
):
    if locals is None:
        locals = {}
    if len(args) != interp.arg_count:
        raise TypeError("Mismatch number of argument types")
    warnings = errors.WarningsFixer(errors.NumbaWarning)

    infer = typeinfer.TypeInferer(typingctx, interp, warnings)
    callstack_ctx = typingctx.callstack.register(targetctx.target, infer, interp.func_id, args)
    # Setup two contexts: 1) callstack setup/teardown 2) flush warnings
    with callstack_ctx, warnings:
        # Seed argument types
        for index, (name, ty) in enumerate(zip(interp.arg_names, args)):
            infer.seed_argument(name, index, ty)

        # Seed return type
        if return_type is not None:
            infer.seed_return(return_type)

        # Seed local types
        for k, v in locals.items():
            infer.seed_type(k, v)

        infer.build_constraint()
        # return errors in case of partial typing
        errs = infer.propagate(raise_errors=raise_errors)
        typemap, restype, calltypes = infer.unify(raise_errors=raise_errors)

    return _TypingResults(typemap, restype, calltypes, errs)


class BaseTypeInference(FunctionPass):
    _raise_errors = True

    def __init__(self):
        FunctionPass.__init__(self)

    def run_pass(self, state):
        """
        Type inference and legalization
        """
        with fallback_context(
            state,
            'Function "%s" failed type inference' % (state.func_id.func_name,),
        ):
            # Type inference
            typemap, return_type, calltypes, errs = type_inference_stage(
                state.typingctx,
                state.targetctx,
                state.func_ir,
                state.args,
                state.return_type,
                state.locals,
                raise_errors=self._raise_errors,
            )
            state.typemap = typemap
            # save errors in case of partial typing
            state.typing_errors = errs
            if self._raise_errors:
                state.return_type = return_type
            state.calltypes = calltypes

        def legalize_return_type(return_type, interp, targetctx):
            """
            Only accept array return type iff it is passed into the function.
            Reject function object return types if in nopython mode.
            """
            if not targetctx.enable_nrt and isinstance(return_type, types.Array):
                # Walk IR to discover all arguments and all return statements
                retstmts = []
                caststmts = {}
                argvars = set()
                for blk in interp.blocks.values():
                    for inst in blk.body:
                        if isinstance(inst, ir.Return):
                            retstmts.append(inst.value.name)
                        elif isinstance(inst, ir.Assign):
                            if isinstance(inst.value, ir.Expr) and inst.value.op == "cast":
                                caststmts[inst.target.name] = inst.value
                            elif isinstance(inst.value, ir.Arg):
                                argvars.add(inst.target.name)

                assert retstmts, "No return statements?"

                for var in retstmts:
                    cast = caststmts.get(var)
                    if cast is None or cast.value.name not in argvars:
                        if self._raise_errors:
                            msg = (
                                "Only accept returning of array passed into "
                                "the function as argument"
                            )
                            raise errors.NumbaTypeError(msg)

            elif isinstance(return_type, types.Function) or isinstance(return_type, types.Phantom):
                if self._raise_errors:
                    msg = "Can't return function object ({}) in nopython mode"
                    raise errors.NumbaTypeError(msg.format(return_type))

        with fallback_context(
            state,
            'Function "%s" has invalid return type' % (state.func_id.func_name,),
        ):
            # We don't need to run legalize_return_type for always inline
            # functions, as they will be inlined into their callers at numba IR
            # stage and never compiled on their own. This allows us to support
            # returning arrays that are not passed in as arguments, which is a
            # common pattern in array-oriented code.
            if state.flags.inline != InlineOptions("always"):
                legalize_return_type(state.return_type, state.func_ir, state.targetctx)
        return True


@register_pass(mutates_CFG=True, analysis_only=False)
class NopythonTypeInference(BaseTypeInference):
    _name = "nopython_type_inference"


@register_pass(mutates_CFG=True, analysis_only=False)
class PartialTypeInference(BaseTypeInference):
    _name = "partial_type_inference"
    _raise_errors = False


@register_pass(mutates_CFG=False, analysis_only=False)
class AnnotateTypes(AnalysisPass):
    _name = "annotate_types"

    def __init__(self):
        AnalysisPass.__init__(self)

    def get_analysis_usage(self, AU):
        AU.add_required(IRLegalization)

    def run_pass(self, state):
        """
        Create type annotation after type inference
        """
        func_ir = state.func_ir.copy()
        state.type_annotation = type_annotations.TypeAnnotation(
            func_ir=func_ir,
            typemap=state.typemap,
            calltypes=state.calltypes,
            lifted=state.lifted,
            lifted_from=state.lifted_from,
            args=state.args,
            return_type=state.return_type,
            html_output=config.HTML,
        )

        if config.ANNOTATE:
            print("ANNOTATION".center(80, "-"))
            print(state.type_annotation)
            print("=" * 80)
        if config.HTML:
            with open(config.HTML, "w") as fout:
                state.type_annotation.html_annotate(fout)

        return False


@register_pass(mutates_CFG=True, analysis_only=False)
class NopythonRewrites(FunctionPass):
    _name = "nopython_rewrites"

    def __init__(self):
        FunctionPass.__init__(self)

    def run_pass(self, state):
        """
        Perform any intermediate representation rewrites after type
        inference.
        """
        # a bunch of these passes are either making assumptions or rely on some
        # very picky and slightly bizarre state particularly in relation to
        # ir.Del presence. To accommodate, ir.Dels are added ahead of running
        # this pass and stripped at the end.

        # Ensure we have an IR and type information.
        assert state.func_ir
        assert isinstance(getattr(state, "typemap", None), dict)
        assert isinstance(getattr(state, "calltypes", None), dict)
        msg = (
            "Internal error in post-inference rewriting "
            "pass encountered during compilation of "
            'function "%s"' % (state.func_id.func_name,)
        )

        pp = postproc.PostProcessor(state.func_ir)
        pp.run(True)
        with fallback_context(state, msg):
            rewrites.rewrite_registry.apply("after-inference", state)
        pp.remove_dels()
        return True


class BaseNativeLowering(abc.ABC, LoweringPass):
    """The base class for a lowering pass. The lowering functionality must be
    specified in inheriting classes by providing an appropriate lowering class
    implementation in the overridden `lowering_class` property."""

    _name = None

    def __init__(self):
        LoweringPass.__init__(self)

    @property
    @abc.abstractmethod
    def lowering_class(self):
        """Returns the class that performs the lowering of the IR describing the
        function that is the target of the current compilation."""
        pass

    def run_pass(self, state):
        if state.library is None:
            codegen = state.targetctx.codegen()
            state.library = codegen.create_library(state.func_id.func_qualname)
            # Enable object caching upfront, so that the library can
            # be later serialized.
            state.library.enable_object_caching()

        library = state.library
        targetctx = state.targetctx
        interp = state.func_ir  # why is it called this?!
        typemap = state.typemap
        restype = state.return_type
        calltypes = state.calltypes
        flags = state.flags
        metadata = state.metadata

        call_conv = state.call_conv
        if call_conv is None:
            call_conv = CUDACallConv(state.targetctx)

        mangler = call_conv.mangler

        msg = "Function %s failed at nopython mode lowering" % (state.func_id.func_name,)
        with fallback_context(state, msg):
            # Lowering
            fndesc = funcdesc.PythonFunctionDescriptor.from_specialized_function(
                interp,
                typemap,
                restype,
                calltypes,
                mangler=mangler,
                inline=flags.forceinline,
                noalias=flags.noalias,
                abi_tags=[flags.get_mangle_string()],
                call_conv=call_conv,
                abi_info=state.abi_info,
            )

            with targetctx.push_code_library(library):
                lower = self.lowering_class(targetctx, library, fndesc, interp, metadata=metadata)
                lower.lower()
                if not flags.no_cpython_wrapper:
                    lower.create_cpython_wrapper(flags.release_gil)

                if not flags.no_cfunc_wrapper:
                    # skip cfunc wrapper generation if unsupported
                    # argument or return types are used
                    for t in state.args:
                        if isinstance(t, (types.Omitted, types.Generator)):
                            break
                    else:
                        if isinstance(restype, (types.Optional, types.Generator)):
                            pass
                        else:
                            lower.create_cfunc_wrapper()

                env = lower.env
                call_helper = lower.call_helper
                del lower

            from numba_cuda_mlir.numba_cuda.compiler import (
                _LowerResult,
            )  # TODO: move this

            if flags.no_compile:
                state["cr"] = _LowerResult(fndesc, call_helper, cfunc=None, env=env)
            else:
                # Prepare for execution
                # Insert native function for use by other jitted-functions.
                # We also register its library to allow for inlining.
                cfunc = targetctx.get_executable(library, fndesc, env)
                targetctx.insert_user_function(cfunc, fndesc, [library])
                state["cr"] = _LowerResult(fndesc, call_helper, cfunc=cfunc, env=env)
        return True


@register_pass(mutates_CFG=True, analysis_only=False)
class NativeLowering(BaseNativeLowering):
    """Lowering pass for a native function IR described solely in terms of
    Numba's standard `numba.cuda.core.ir` nodes."""

    _name = "native_lowering"

    @property
    def lowering_class(self):
        # The vendored llvmlite Lower is dead on the MLIR path (kernels are
        # lowered by MLIRLower, not by this pass). Returning None so the dead
        # Lower class can be removed; if this pass ever ran it would fail loudly.
        return None


@register_pass(mutates_CFG=False, analysis_only=True)
class NoPythonSupportedFeatureValidation(AnalysisPass):
    """NoPython Mode check: Validates the IR to ensure that features in use are
    in a form that is supported"""

    _name = "nopython_supported_feature_validation"

    def __init__(self):
        AnalysisPass.__init__(self)

    def run_pass(self, state):
        raise_on_unsupported_feature(state.func_ir, state.typemap)
        warn_deprecated(state.func_ir, state.typemap)
        return False


@register_pass(mutates_CFG=False, analysis_only=True)
class IRLegalization(AnalysisPass):
    _name = "ir_legalization"

    def __init__(self):
        AnalysisPass.__init__(self)

    def run_pass(self, state):
        # NOTE: this function call must go last, it checks and fixes invalid IR!
        check_and_legalize_ir(state.func_ir, flags=state.flags)
        return True


@register_pass(mutates_CFG=True, analysis_only=False)
class NoPythonBackend(LoweringPass):
    _name = "nopython_backend"

    def __init__(self):
        LoweringPass.__init__(self)

    def run_pass(self, state):
        """
        Back-end: Generate LLVM IR from Numba IR, compile to machine code
        """
        lowered = state["cr"]
        signature = typing.signature(state.return_type, *state.args)

        from numba_cuda_mlir.numba_cuda.compiler import cuda_compile_result

        state.cr = cuda_compile_result(
            typing_context=state.typingctx,
            target_context=state.targetctx,
            entry_point=lowered.cfunc,
            typing_error=state.status.fail_reason,
            type_annotation=state.type_annotation,
            library=state.library,
            call_helper=lowered.call_helper,
            signature=signature,
            objectmode=False,
            lifted=state.lifted,
            fndesc=lowered.fndesc,
            environment=lowered.env,
            metadata=state.metadata,
            reload_init=state.reload_init,
        )
        return True


@register_pass(mutates_CFG=True, analysis_only=False)
class InlineOverloads(FunctionPass):
    """
    This pass will inline a function wrapped by the numba.cuda.extending.overload
    decorator directly into the site of its call depending on the value set in
    the 'inline' kwarg to the decorator.

    This is a typed pass. CFG simplification and DCE are performed on
    completion.
    """

    _name = "inline_overloads"

    def __init__(self):
        FunctionPass.__init__(self)

    _DEBUG = False

    def run_pass(self, state):
        """Run inlining of overloads"""
        if self._DEBUG:
            print("before overload inline".center(80, "-"))
            print(state.func_id.unique_name)
            print(state.func_ir.dump())
            print("".center(80, "-"))
        from numba_cuda_mlir.numba_cuda.core.inline_closurecall import (
            InlineWorker,
            callee_ir_validator,
        )

        inline_worker = InlineWorker(
            state.typingctx,
            state.targetctx,
            state.locals,
            state.pipeline,
            state.flags,
            callee_ir_validator,
            state.typemap,
            state.calltypes,
        )
        modified = False
        work_list = list(state.func_ir.blocks.items())
        # use a work list, look for call sites via `ir.Expr.op == call` and
        # then pass these to `self._do_work` to make decisions about inlining.
        while work_list:
            label, block = work_list.pop()
            for i, instr in enumerate(block.body):
                # TO-DO: other statements (setitem)
                if isinstance(instr, ir.Assign):
                    expr = instr.value
                    if isinstance(expr, ir.Expr):
                        workfn = self._do_work_expr

                        if guard(
                            workfn,
                            state,
                            work_list,
                            block,
                            i,
                            expr,
                            inline_worker,
                        ):
                            modified = True
                            break  # because block structure changed

        if self._DEBUG:
            print("after overload inline".center(80, "-"))
            print(state.func_id.unique_name)
            print(state.func_ir.dump())
            print("".center(80, "-"))

        if modified:
            # Remove dead blocks, this is safe as it relies on the CFG only.
            cfg = compute_cfg_from_blocks(state.func_ir.blocks)
            for dead in cfg.dead_nodes():
                del state.func_ir.blocks[dead]
            # clean up blocks
            dead_code_elimination(state.func_ir, typemap=state.typemap)
            # clean up unconditional branches that appear due to inlined
            # functions introducing blocks
            state.func_ir.blocks = simplify_CFG(state.func_ir.blocks)

        if self._DEBUG:
            print("after overload inline DCE".center(80, "-"))
            print(state.func_id.unique_name)
            print(state.func_ir.dump())
            print("".center(80, "-"))
        return True

    def _get_attr_info(self, state, expr):
        recv_type = state.typemap[expr.value.name]
        recv_type = types.unliteral(recv_type)
        matched = state.typingctx.find_matching_getattr_template(
            recv_type,
            expr.attr,
        )
        if not matched:
            return None

        template = matched["template"]
        if getattr(template, "is_method", False):
            # The attribute template is representing a method.
            # Don't inline the getattr.
            return None

        templates = [template]
        sig = typing.signature(matched["return_type"], recv_type)
        arg_typs = sig.args
        is_method = False

        return templates, sig, arg_typs, is_method

    def _get_callable_info(self, state, expr):
        def get_func_type(state, expr):
            func_ty = None
            if expr.op == "call":
                # check this is a known and typed function
                try:
                    func_ty = state.typemap[expr.func.name]
                except KeyError:
                    # e.g. Calls to CUDA Intrinsic have no mapped type
                    # so KeyError
                    return None
                if not hasattr(func_ty, "get_call_type"):
                    return None

            elif is_operator_or_getitem(expr):
                func_ty = state.typingctx.resolve_value_type(expr.fn)
            else:
                return None

            return func_ty

        if expr.op == "call":
            # try and get a definition for the call, this isn't always
            # possible as it might be a eval(str)/part generated
            # awaiting update etc. (parfors)
            to_inline = None
            try:
                to_inline = state.func_ir.get_definition(expr.func)
            except Exception:
                return None

            # do not handle closure inlining here, another pass deals with that
            if getattr(to_inline, "op", False) == "make_function":
                return None

        func_ty = get_func_type(state, expr)
        if func_ty is None:
            return None

        sig = state.calltypes[expr]
        if not sig:
            return None

        templates, arg_typs, is_method = None, None, False
        if getattr(func_ty, "template", None) is not None:
            # @overload_method
            is_method = True
            templates = [func_ty.template]
            arg_typs = (func_ty.template.this,) + sig.args
        else:
            # @overload case
            templates = getattr(func_ty, "templates", None)
            arg_typs = sig.args

        return templates, sig, arg_typs, is_method

    def _do_work_expr(self, state, work_list, block, i, expr, inline_worker):
        def select_template(templates, args):
            if templates is None:
                return None

            impl = None
            for template in templates:
                inline_type = getattr(template, "_inline", None)
                if inline_type is None:
                    # inline not defined
                    continue
                if args not in template._inline_overloads:
                    # skip overloads not matching signature
                    continue
                if not inline_type.is_never_inline:
                    try:
                        impl = template._overload_func(*args)
                        if impl is None:
                            raise Exception  # abort for this template
                        break
                    except Exception:
                        continue
            else:
                return None

            return template, inline_type, impl

        inlinee_info = None
        if expr.op == "getattr":
            inlinee_info = self._get_attr_info(state, expr)
        else:
            inlinee_info = self._get_callable_info(state, expr)

        if not inlinee_info:
            return False

        templates, sig, arg_typs, is_method = inlinee_info
        inlinee = select_template(templates, arg_typs)
        if inlinee is None:
            return False
        template, inlinee_type, impl = inlinee

        return self._run_inliner(
            state,
            inlinee_type,
            sig,
            template,
            arg_typs,
            expr,
            i,
            impl,
            block,
            work_list,
            is_method,
            inline_worker,
        )

    def _run_inliner(
        self,
        state,
        inline_type,
        sig,
        template,
        arg_typs,
        expr,
        i,
        impl,
        block,
        work_list,
        is_method,
        inline_worker,
    ):
        do_inline = True
        if not inline_type.is_always_inline:
            from numba_cuda_mlir.numba_cuda.typing.templates import _inline_info

            caller_inline_info = _inline_info(state.func_ir, state.typemap, state.calltypes, sig)

            # must be a cost-model function, run the function
            iinfo = template._inline_overloads[arg_typs]["iinfo"]
            if inline_type.has_cost_model:
                do_inline = inline_type.value(expr, caller_inline_info, iinfo)
            else:
                assert "unreachable"

        if do_inline:
            if is_method:
                if not self._add_method_self_arg(state, expr):
                    return False
            arg_typs = template._inline_overloads[arg_typs]["folded_args"]
            iinfo = template._inline_overloads[arg_typs]["iinfo"]
            freevars = iinfo.func_ir.func_id.func.__code__.co_freevars
            _, _, _, new_blocks = inline_worker.inline_ir(
                state.func_ir,
                block,
                i,
                iinfo.func_ir,
                freevars,
                arg_typs=arg_typs,
            )
            if work_list is not None:
                for blk in new_blocks:
                    work_list.append(blk)
            return True
        else:
            return False

    def _add_method_self_arg(self, state, expr):
        func_def = guard(get_definition, state.func_ir, expr.func)
        if func_def is None:
            return False
        expr.args.insert(0, func_def.value)
        return True


@register_pass(mutates_CFG=False, analysis_only=False)
class DeadCodeElimination(FunctionPass):
    """
    Does dead code elimination
    """

    _name = "dead_code_elimination"

    def __init__(self):
        FunctionPass.__init__(self)

    def run_pass(self, state):
        dead_code_elimination(state.func_ir, state.typemap)
        return True


@register_pass(mutates_CFG=False, analysis_only=False)
class PreLowerStripPhis(FunctionPass):
    """Remove phi nodes (ir.Expr.phi) introduced by SSA.

    This is needed before Lowering because the phi nodes in Numba IR do not
    match the semantics of phi nodes in LLVM IR. In Numba IR, phi nodes may
    expand into multiple LLVM instructions.
    """

    _name = "strip_phis"

    def __init__(self):
        FunctionPass.__init__(self)

    def run_pass(self, state):
        state.func_ir = self._strip_phi_nodes(state.func_ir)

        # Rerun postprocessor to update metadata
        post_proc = postproc.PostProcessor(state.func_ir)
        post_proc.run(emit_dels=False)

        # Ensure we are not in objectmode generator
        if state.func_ir.generator_info is not None and state.typemap is not None:
            # Rebuild generator type
            # TODO: move this into PostProcessor
            gentype = state.return_type
            state_vars = state.func_ir.generator_info.state_vars
            state_types = [state.typemap[k] for k in state_vars]
            state.return_type = types.Generator(
                gen_func=gentype.gen_func,
                yield_type=gentype.yield_type,
                arg_types=gentype.arg_types,
                state_types=state_types,
                has_finalizer=gentype.has_finalizer,
            )
        return True

    def _strip_phi_nodes(self, func_ir):
        """Strip Phi nodes from ``func_ir``

        For each phi node, put incoming value to their respective incoming
        basic-block at possibly the latest position (i.e. after the latest
        assignment to the corresponding variable).
        """
        exporters = defaultdict(list)
        phis = set()
        # Find all variables that needs to be exported
        for block in func_ir.blocks.values():
            for assign in block.find_insts(ir.Assign):
                if isinstance(assign.value, ir.Expr):
                    if assign.value.op == "phi":
                        phis.add(assign)
                        phi = assign.value
                        for ib, iv in zip(phi.incoming_blocks, phi.incoming_values):
                            exporters[ib].append((assign.target, iv))

        # Rewrite the blocks with the new exporting assignments
        newblocks = {}
        for label, block in func_ir.blocks.items():
            newblk = copy(block)
            newblocks[label] = newblk

            # strip phis
            newblk.body = [stmt for stmt in block.body if stmt not in phis]

            # insert exporters
            # Get the block's terminator location for use when the variable
            # was not assigned in this block (e.g., in continue/break paths).
            term_loc = block.terminator.loc if block.is_terminated else block.loc
            for target, rhs in exporters[label]:
                # If RHS is undefined
                if rhs is ir.UNDEFINED:
                    # Put in a NULL initializer, set the location to be in what
                    # will eventually materialize as the prologue.
                    rhs = ir.Expr.null(loc=func_ir.loc)

                # Insert at the earliest possible location; i.e. after the
                # last assignment to rhs
                assignments = [stmt for stmt in newblk.find_insts(ir.Assign) if stmt.target == rhs]
                if assignments:
                    # Variable was assigned in this block; use RHS location
                    # to keep the location info at the assignment site.
                    assign = ir.Assign(target=target, value=rhs, loc=rhs.loc)
                    last_assignment = assignments[-1]
                    newblk.insert_after(assign, last_assignment)
                else:
                    # Variable was NOT assigned in this block (just passed
                    # through, e.g., in continue/break blocks). Use the
                    # terminator location to preserve location info.
                    assign = ir.Assign(target=target, value=rhs, loc=term_loc)
                    newblk.prepend(assign)

        func_ir.blocks = newblocks
        return func_ir



# SPDX-FileCopyrightText: Copyright (c) 2026 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
from dataclasses import dataclass
from numba_cuda_mlir.errors import InternalCompilerError
import math as builtin_math
import operator
from numba_cuda_mlir.lowering_utilities import (
    DeferredMethodCall,
    type_conversions,
    convert,
    constant,
    numpy_implicit_type_promotion,
    memref_to_tensor,
    memref_to_value_tensor,
    value_tensor_to_storage_memref,
    broadcast_shapes_for_binary_op,
    index_of,
    tensor_to_memref,
    simple_scalar_conversion_op,
    expensive_coerce_tensor_type,
    try_extract_constant,
)
from numba_cuda_mlir.logging import trace
from numba_cuda_mlir.mlir_lowering import MLIRLower
from numba_cuda_mlir import types
from numba_cuda_mlir.numba_cuda.core.errors import TypingError
from numba_cuda_mlir.numba_cuda.extending import intrinsic
from numba_cuda_mlir._mlir.extras import types as T
from numba_cuda_mlir._mlir.dialects import (
    arith,
    math,
    nvvm,
)
from numba_cuda_mlir.mlir.dialect_exts.math import ipowi
from numba_cuda_mlir.lowering_registry import LoweringRegistry

registry = LoweringRegistry()
lower = registry.lower
lower_getattr = registry.lower_getattr
from numba_cuda_mlir.lowering_utilities import (
    RangeObject,
    int_of,
)
from numba_cuda_mlir._mlir import ir
from .ufunc_registry import UFuncRegistry

ufunc_registry = UFuncRegistry("builtins")


def _get_range_object(builder, args: list[ir.Value]) -> tuple[ir.Value, ir.Value, ir.Value]:
    match args:
        case [stop]:
            stop_type = builder.get_mlir_type(stop)
            return (
                int_of(0, ty=stop_type),
                int_of(builder.load_var(stop), ty=stop_type, signed=True),
                int_of(1, ty=stop_type),
            )
        case [start, stop]:
            start_type = builder.get_mlir_type(start)
            stop_type = builder.get_mlir_type(stop)
            return (
                int_of(builder.load_var(start), ty=start_type, signed=True),
                int_of(builder.load_var(stop), ty=stop_type, signed=True),
                int_of(1, ty=stop_type),
            )
        case [start, stop, step]:
            return (
                int_of(
                    builder.load_var(start),
                    ty=builder.get_mlir_type(start),
                    signed=True,
                ),
                int_of(builder.load_var(stop), ty=builder.get_mlir_type(stop), signed=True),
                int_of(builder.load_var(step), ty=builder.get_mlir_type(step), signed=True),
            )
        case _:
            raise ValueError(f"Invalid arguments for range: {args}")


@lower(operator.mul, types.UniTuple, types.Literal)
@lower(operator.mul, types.Tuple, types.Literal)
@lower(operator.mul, types.Literal, types.UniTuple)
def lower_tuple_multiply(builder, target, args, kwargs):
    trace()
    assert not kwargs and len(args) == 2, "tuple multiply expects 2 arguments"
    tup, how_many = builder.load_vars(args)
    assert isinstance(tup, tuple), f"Expected tuple, got {type(tup)}"
    match tup, how_many:
        case tuple(), (int() | ir.Value()):
            pass
        case ((int() | ir.Value()), tuple()):
            tup, how_many = how_many, tup
        case _:
            raise TypeError(f"Invalid arguments for tuple multiply: {tup=} {how_many=}")

    if how_many := try_extract_constant(how_many):
        builder.store_var(target, tup * how_many)
        return

    raise InternalCompilerError(f"Expected literal integer, got {type(how_many)}")


@lower(tuple, types.UniTuple)
@lower(tuple, types.Tuple)
def lower_tuple_builtin(builder, target, args, kwargs):
    trace()
    assert not kwargs and len(args) == 1, "tuple() expects 1 argument"
    value = builder.load_var(args[0])
    if not isinstance(value, tuple):
        value = (value,)
    builder.store_var(target, value)


@lower(operator.add, types.UniTuple, types.UniTuple)
@lower(operator.add, types.Tuple, types.Tuple)
@lower(operator.add, types.UniTuple, types.Tuple)
@lower(operator.add, types.Tuple, types.UniTuple)
def lower_tuple_concat(builder, target, args, kwargs):
    trace()
    assert not kwargs and len(args) == 2, "tuple concat expects 2 arguments"
    lhs, rhs = builder.load_vars(args)
    if not isinstance(lhs, tuple):
        lhs = (lhs,)
    if not isinstance(rhs, tuple):
        rhs = (rhs,)
    builder.store_var(target, lhs + rhs)


@lower(range, types.Number)
@lower(range, types.Number, types.Number)
@lower(range, types.Number, types.Number, types.Number)
def lower_range(builder: MLIRLower, target, args, kwargs):
    start, stop, step = _get_range_object(builder, args)
    ro = RangeObject(builder, start, stop, step)
    builder.store_var(target, ro)


@lower(operator.contains, types.Tuple, types.Number)
def tuple_contains_cg(builder, target, args, kwargs):
    tup = builder.load_var(args[0])
    item = builder.load_var(args[1])
    from numba_cuda_mlir.lowering_utilities import bool_of

    # Extract constant values from the tuple elements
    constant_values = []
    for x in tup:
        const_val = try_extract_constant(x)
        if const_val is None:
            raise NotImplementedError(
                f"Tuple contains is not implemented for non-constant values: {x}"
            )
        constant_values.append(const_val)

    # Check if item is a constant and can be checked at compile time
    item_const = try_extract_constant(item)
    if item_const is not None:
        builder.store_var(target, bool_of(item_const in constant_values))
    else:
        # Runtime check - item is not a constant
        from numba_cuda_mlir.lowering_utilities import false, equal

        result = false()
        for const_val in constant_values:
            result = arith.ori(result, equal(item, int_of(const_val, item.type)))
        builder.store_var(target, result)


@lower(operator.contains, types.UniTuple, types.Number)
def unituple_contains_cg(builder, target, args, kwargs):
    trace("args=%s", args)
    from numba_cuda_mlir._mlir.dialects import linalg, tensor
    from numba_cuda_mlir.lowering_utilities import (
        false,
        equal,
        bool_of,
        concretize_tuple_to_tensor,
    )

    tup: tuple = builder.load_var(args[0])
    assert isinstance(tup, tuple), f"Expected Python tuple, got {type(tup)}"
    tup = concretize_tuple_to_tensor(tup)

    item: ir.Value = builder.load_var(args[1])

    mr_type = tup.type
    rank = mr_type.rank
    if not mr_type.has_rank:
        raise NotImplementedError("NYI: unranked memrefs")
    if mr_type.rank != 1:
        raise TypeError(f"Expected a 1-dimensional memref, got a {rank}-dimensional memref")

    def body(op: linalg.ReduceOp, element: ir.Value, accumulator: ir.Value):
        found = equal(element, item)
        found = arith.ori(found, accumulator)
        linalg.yield_([found])

    result_type = ir.RankedTensorType.get((), T.bool())
    init = tensor.splat(result_type, false(), [])
    dims_attr = ir.DenseI64ArrayAttr.get([0])
    op = linalg.ReduceOp(
        result=[result_type],
        inputs=[tup],
        inits=[init],
        dimensions=dims_attr,
    )
    block = op.combiner.blocks.append(tup.type.element_type, result_type.element_type)
    with ir.InsertionPoint(block):
        body(op, *block.arguments)
    result = tensor.extract(op.results[0], [])
    builder.store_var(target, bool_of(result))


for exc_type in (
    ArithmeticError,
    AssertionError,
    AttributeError,
    BufferError,
    EOFError,
    ImportError,
    LookupError,
    MemoryError,
    NameError,
    OSError,
    ReferenceError,
    RuntimeError,
    SystemError,
    TypeError,
    ValueError,
):

    @lower(exc_type, types.StringLiteral)
    def lower_exc_type(builder, target, args, kwargs, exc_type=exc_type):
        message = builder.load_var(args[0])
        builder.store_var(target, exc_type(message))


@lower(operator.eq, types.StringLiteral, types.StringLiteral)
def operator_eq_string_literal_lower(builder, target, args, kwargs):
    """Lower str == str for string literals (constant-fold at compile time)."""
    left_type = builder.get_numba_type(args[0].name)
    right_type = builder.get_numba_type(args[1].name)
    result = arith.constant(
        result=T.bool(),
        value=(left_type.literal_value == right_type.literal_value),
    )
    builder.store_var(target, result)


@lower(operator.not_, types.Number)
@lower(operator.not_, types.Boolean)
def lower_not(builder, target, args, kwargs):
    """Boolean negation: not x == (x == 0)."""
    import numba_cuda_mlir.lowering_utilities

    operand = builder.load_var(args[0])
    c0 = numba_cuda_mlir.lowering_utilities.constant(0, operand.type)
    res = numba_cuda_mlir.lowering_utilities.equal(operand, c0)
    builder.store_var(target, res)


def lower_broadcasted_binary(builder, target, args, kwargs):
    from numba_cuda_mlir._mlir.dialects import linalg, shape, tensor
    from numba_cuda_mlir.lowering_utilities import index_of, tensor_to_memref

    arg_vars = args
    op = builder.load_var(arg_vars[0])
    lhs = builder.load_var(arg_vars[1])
    rhs = builder.load_var(arg_vars[2])
    lhs_type = builder.get_numba_type(arg_vars[1].name)
    rhs_type = builder.get_numba_type(arg_vars[2].name)
    target_type = builder.get_numba_type(target.name)
    result_element_type = builder.get_value_type(target_type.dtype)

    if isinstance(lhs_type, types.Array):
        lhs = memref_to_value_tensor(lhs_type, lhs)
    if isinstance(rhs_type, types.Array):
        rhs = memref_to_value_tensor(rhs_type, rhs)

    lhs, rhs = broadcast_shapes_for_binary_op(lhs, rhs, builder)
    lhs = expensive_coerce_tensor_type(lhs, result_element_type)
    rhs = expensive_coerce_tensor_type(rhs, result_element_type)

    sh = shape.shape_of(lhs)
    dims = [shape.get_extent(sh, index_of(i)) for i in range(lhs.type.rank)]
    dtype = T.tensor(*lhs.type.shape, result_element_type)
    empty = tensor.empty(sizes=dims, element_type=result_element_type)

    lhs_conv, rhs_conv = (
        simple_scalar_conversion_op(lhs.type.element_type, result_element_type),
        simple_scalar_conversion_op(rhs.type.element_type, result_element_type),
    )
    lhs_etype, rhs_etype = lhs.type.element_type, rhs.type.element_type

    @linalg.map(
        result=[dtype],
        inputs=[lhs, rhs],
        init=empty,
    )
    def binop(lhs: lhs_etype, rhs: rhs_etype, init: result_element_type):
        lhs, rhs = lhs_conv(lhs), rhs_conv(rhs)
        match op:
            case operator.sub:
                return (
                    arith.subi(lhs, rhs)
                    if isinstance(result_element_type, ir.IntegerType)
                    else arith.subf(lhs, rhs)
                )
            case operator.add:
                return (
                    arith.addi(lhs, rhs)
                    if isinstance(result_element_type, ir.IntegerType)
                    else arith.addf(lhs, rhs)
                )
            case operator.mul:
                return (
                    arith.muli(lhs, rhs)
                    if isinstance(result_element_type, ir.IntegerType)
                    else arith.mulf(lhs, rhs)
                )
            case operator.truediv | operator.itruediv:
                return arith.divf(lhs, rhs)
            case operator.floordiv | operator.ifloordiv:
                return arith.divsi(lhs, rhs)
            case operator.pow | builtin_math.pow:
                match lhs.type, rhs.type:
                    case ir.FloatType(), ir.FloatType():
                        return math.powf(lhs, rhs)
                    case ir.FloatType(), ir.IntegerType():
                        return math.fpowi(lhs, rhs)
                    case ir.IntegerType(), ir.FloatType():
                        raise InternalCompilerError("NYI: integer power of float, unreachable")
                    case ir.IntegerType(), ir.IntegerType():
                        rhs = convert(rhs, lhs.type)
                        return ipowi(lhs, rhs)
            case _:
                raise NotImplementedError(f"Not implemented for operator {op}")

    result = value_tensor_to_storage_memref(target_type, binop)
    builder.store_var(target, result)


@ufunc_registry.register(operator.floordiv)
@lower(operator.floordiv, types.Array, types.Array)
def lower_broadcasted_floor_division(builder, target, args, kwargs):
    from numba_cuda_mlir._mlir.dialects import linalg, shape, tensor

    trace()

    target_type = builder.get_numba_type(target.name)
    result_value_type = builder.get_value_type(target_type.dtype)
    compute_type = result_value_type
    if isinstance(result_value_type, ir.FloatType):
        compute_type = type_conversions.integer_of_width(result_value_type.width)
        compute_type = type_conversions.to_mlir_type(compute_type)

    lhs_arg, rhs_arg = args
    lhs_type = builder.get_numba_type(lhs_arg.name)
    rhs_type = builder.get_numba_type(rhs_arg.name)
    lhs = builder.load_var(lhs_arg)
    rhs = builder.load_var(rhs_arg)
    if isinstance(lhs_type, types.Array):
        lhs = memref_to_value_tensor(lhs_type, lhs)
    if isinstance(rhs_type, types.Array):
        rhs = memref_to_value_tensor(rhs_type, rhs)

    lhs, rhs = broadcast_shapes_for_binary_op(lhs, rhs, builder)
    lhs = expensive_coerce_tensor_type(lhs, compute_type)
    rhs = expensive_coerce_tensor_type(rhs, compute_type)
    sh = shape.shape_of(lhs)
    dims = [shape.get_extent(sh, index_of(i)) for i in range(lhs.type.rank)]
    result_type = T.tensor(*lhs.type.shape, result_value_type)
    empty = tensor.empty(sizes=dims, element_type=result_value_type)

    lhs_conv, rhs_conv = (
        simple_scalar_conversion_op(lhs.type.element_type, compute_type),
        simple_scalar_conversion_op(rhs.type.element_type, compute_type),
    )
    lhs_etype, rhs_etype = lhs.type.element_type, rhs.type.element_type

    @linalg.map(
        result=[result_type],
        inputs=[lhs, rhs],
        init=empty,
    )
    def casted_div(lhs: lhs_etype, rhs: rhs_etype, init: result_value_type):
        lhs, rhs = lhs_conv(lhs), rhs_conv(rhs)
        result = arith.divsi(lhs, rhs)
        return convert(result, result_value_type)

    result = value_tensor_to_storage_memref(target_type, casted_div)
    builder.store_var(target, result)


@ufunc_registry.register(operator.truediv)
@lower(operator.truediv, types.Array, types.Array)
def lower_broadcasted_div(builder, target, args, kwargs):
    from numba_cuda_mlir._mlir.dialects import linalg, shape, tensor

    trace()

    target_type = builder.get_numba_type(target.name)
    result_value_type = builder.get_value_type(target_type.dtype)
    assert isinstance(result_value_type, ir.FloatType)

    lhs_arg, rhs_arg = args
    lhs_type = builder.get_numba_type(lhs_arg.name)
    rhs_type = builder.get_numba_type(rhs_arg.name)
    lhs = builder.load_var(lhs_arg)
    rhs = builder.load_var(rhs_arg)
    if isinstance(lhs_type, types.Array):
        lhs = memref_to_value_tensor(lhs_type, lhs)
    if isinstance(rhs_type, types.Array):
        rhs = memref_to_value_tensor(rhs_type, rhs)

    lhs, rhs = broadcast_shapes_for_binary_op(lhs, rhs, builder)
    lhs = expensive_coerce_tensor_type(lhs, result_value_type)
    rhs = expensive_coerce_tensor_type(rhs, result_value_type)
    sh = shape.shape_of(lhs)
    dims = [shape.get_extent(sh, index_of(i)) for i in range(lhs.type.rank)]
    result_type = T.tensor(*lhs.type.shape, result_value_type)
    empty = tensor.empty(sizes=dims, element_type=result_value_type)

    lhs_conv, rhs_conv = (
        simple_scalar_conversion_op(lhs.type.element_type, result_value_type),
        simple_scalar_conversion_op(rhs.type.element_type, result_value_type),
    )
    lhs_etype, rhs_etype = lhs.type.element_type, rhs.type.element_type

    @linalg.map(
        result=[result_type],
        inputs=[lhs, rhs],
        init=empty,
    )
    def casted_div(lhs: lhs_etype, rhs: rhs_etype, init: result_value_type):
        lhs, rhs = lhs_conv(lhs), rhs_conv(rhs)
        return arith.divf(lhs, rhs)

    result = value_tensor_to_storage_memref(target_type, casted_div)
    builder.store_var(target, result)


@lower(int, types.Number)
@lower(int, types.Boolean)
@lower(bool, types.Number)
@lower(bool, types.Boolean)
@lower(float, types.Number)
@lower(float, types.Boolean)
def type_convert(builder, target, args, kwargs):
    target_numba_ty = builder.get_numba_type(target)
    to_type = builder.get_mlir_type(target)
    to_signed = isinstance(target_numba_ty, types.Integer) and target_numba_ty.signed
    source_numba_ty = builder.get_numba_type(args[0])

    if isinstance(target_numba_ty, (types.IntegerLiteral, types.Literal)):
        result = constant(target_numba_ty.literal_value, to_type)
        builder.store_var(target, result)
        return

    value = builder.load_var(args[0])

    if (const_val := try_extract_constant(value)) is not None:
        result = constant(const_val, to_type)
        builder.store_var(target, result)
        return

    if isinstance(value.type, ir.BF16Type) and isinstance(to_type, ir.IntegerType):
        value = (
            arith.fptosi(out=to_type, in_=value)
            if to_type.width > 1
            else arith.fptoui(out=to_type, in_=value)
        )
    else:
        value = convert(
            value,
            to_type,
            signed=to_signed and not isinstance(source_numba_ty, types.Boolean),
        )
    builder.store_var(target, value)


@lower(float, types.StringLiteral)
def float_from_string_literal(builder, target, args, kwargs):
    """Lower float("nan"), float("inf"), float("-inf"), etc."""
    from numba_cuda_mlir._mlir.extras import types as T

    # Get the string literal value from the type
    arg_type = builder.get_numba_type(args[0])
    string_val = arg_type.literal_value.lower()

    if string_val == "nan":
        val = float("nan")
    elif string_val == "inf" or string_val == "infinity":
        val = float("inf")
    elif string_val == "-inf" or string_val == "-infinity":
        val = float("-inf")
    else:
        raise NotImplementedError(f"float() with string literal '{string_val}' not supported")

    result = arith.constant(result=T.f64(), value=val)
    builder.store_var(target, result)


for type_name in dir(types):
    typ = getattr(types, type_name)
    if isinstance(typ, types.Type):
        lower(typ, types.Number)(type_convert)

# Register numpy type constructors (e.g. np.float32(x), np.int64(x))
import numpy as np

_numpy_type_constructors = [
    np.int8,
    np.int16,
    np.int32,
    np.int64,
    np.uint8,
    np.uint16,
    np.uint32,
    np.uint64,
    np.float16,
    np.float32,
    np.float64,
    np.bool_,
    np.complex64,
    np.complex128,
]
for np_type in _numpy_type_constructors:
    lower(np_type, types.Number)(type_convert)

# When callee is a type (e.g. uint8(x)), lookup uses types.NumberClass
lower(types.NumberClass, types.Number)(type_convert)
lower(types.NumberClass, types.Boolean)(type_convert)


@lower(max, types.Number, types.Number)
def lower_max(builder, target, args, kwargs):
    """Lower built-in max(a, b) for numeric types."""
    from numba_cuda_mlir.lowering_utilities import coerce_numpy_scalars_for_binary_op

    a = builder.load_var(args[0])
    b = builder.load_var(args[1])

    # Promote types using numpy implicit type promotion
    a, b = coerce_numpy_scalars_for_binary_op(a, b)

    # Use appropriate max operation based on type
    if isinstance(a.type, ir.FloatType):
        result = arith.maximumf(a, b)
    elif isinstance(a.type, ir.IntegerType):
        # For simplicity, use signed max (could enhance to detect signed/unsigned)
        result = arith.maxsi(a, b)
    else:
        raise NotImplementedError(f"max not implemented for type {a.type}")

    builder.store_var(target, result)


@lower(min, types.Number, types.Number)
def lower_min(builder, target, args, kwargs):
    """Lower built-in min(a, b) for numeric types."""
    from numba_cuda_mlir.lowering_utilities import coerce_numpy_scalars_for_binary_op

    a = builder.load_var(args[0])
    b = builder.load_var(args[1])

    # Promote types using numpy implicit type promotion
    a, b = coerce_numpy_scalars_for_binary_op(a, b)

    # Use appropriate min operation based on type
    if isinstance(a.type, ir.FloatType):
        result = arith.minimumf(a, b)
    elif isinstance(a.type, ir.IntegerType):
        # For simplicity, use signed min (could enhance to detect signed/unsigned)
        result = arith.minsi(a, b)
    else:
        raise NotImplementedError(f"min not implemented for type {a.type}")

    builder.store_var(target, result)


@lower(round, types.Number, types.Number)
def lower_round(builder, target, args, kwargs):
    from numba_cuda_mlir.runtime import round

    trace()
    value = builder.load_var(args[0])
    ndigits = builder.load_var(args[1])
    fname = f"round_ndigits_type_{str(value.type)}_type_{str(ndigits.type)}"
    if func := getattr(round, fname, None):
        builder.lower_call_external_mlir_library_function(target, func, args, {})
        return
    raise NotImplementedError(f"round intrinsic for types {value.type} and {ndigits.type}")


@lower(round, types.Number)
def lower_round_single(builder, target, args, kwargs):
    from numba_cuda_mlir._mlir.dialects import math

    value = builder.load_var(args[0])
    res = math.round(value)
    builder.store_var(target, res)


@lower(breakpoint)
def lower_breakpoint(builder, target, args, kwargs):
    nvvm.breakpoint()
    builder.store_var(target, None)


@lower(abs, types.Number)
def lower_abs_number(builder, target, args, kwargs):
    """Lower built-in abs() for numeric types (int, float, complex)."""
    assert len(args) == 1, "abs expects 1 argument"
    value = builder.load_var(args[0])
    arg_type = builder.get_numba_type(args[0].name)

    if isinstance(arg_type, types.Complex):
        from numba_cuda_mlir._mlir.dialects import complex as complex_dialect

        result = complex_dialect.abs(value)
    elif isinstance(arg_type, types.Float):
        result = math.absf(value)
    elif isinstance(arg_type, types.Integer):
        if arg_type.signed:
            result = math.absi(value)
        else:
            result = value
    else:
        result = math.absi(value)

    builder.store_var(target, result)


@lower_getattr(types.Number, "bit_count")
def lower_number_bit_count(_, builder, target, num):
    def lower(builder, target, args, kwargs):
        from numba_cuda_mlir._mlir.dialects import math

        value = builder.load_var(num)
        res = math.ctpop(value)
        builder.store_var(target, res)

    builder.store_var(target, DeferredMethodCall(num, lower))


@lower(operator.add, types.Array, types.Array)
@lower(operator.add, types.Array, types.Number)
@lower(operator.add, types.Number, types.Array)
def operator_add_array_lower(builder, target, args, kwargs):
    """Lower operator.add for arrays by using linalg.add"""
    from numba_cuda_mlir._mlir.dialects import linalg
    from numba_cuda_mlir.lowering_utilities.linalg_lowering import lower_np_binop

    target_type = builder.get_numba_type(target.name)
    lower_np_binop(builder, target, target_type, args, linalg.add)


@lower(operator.iadd, types.Array, types.Array)
@lower(operator.iadd, types.Array, types.Number)
@lower(operator.iadd, types.Number, types.Array)
def operator_iadd_array_lower(builder, target, args, kwargs):
    """Lower operator.iadd for arrays by using linalg.add"""
    from numba_cuda_mlir._mlir.dialects import linalg
    from numba_cuda_mlir.lowering_utilities.linalg_lowering import lower_np_binop

    target_type = builder.get_numba_type(target.name)
    lower_np_binop(builder, target, target_type, args, linalg.add)


@lower(operator.sub, types.Array, types.Array)
@lower(operator.sub, types.Array, types.Number)
@lower(operator.sub, types.Number, types.Array)
def operator_sub_array_lower(builder, target, args, kwargs):
    """Lower operator.sub for arrays by using linalg.sub"""
    from numba_cuda_mlir._mlir.dialects import linalg
    from numba_cuda_mlir.lowering_utilities.linalg_lowering import lower_np_binop

    target_type = builder.get_numba_type(target.name)
    lower_np_binop(builder, target, target_type, args, linalg.sub)


@lower(operator.isub, types.Array, types.Array)
@lower(operator.isub, types.Array, types.Number)
@lower(operator.isub, types.Number, types.Array)
def operator_isub_array_lower(builder, target, args, kwargs):
    """Lower operator.isub for arrays by using linalg.sub"""
    from numba_cuda_mlir._mlir.dialects import linalg
    from numba_cuda_mlir.lowering_utilities.linalg_lowering import lower_np_binop

    target_type = builder.get_numba_type(target.name)
    lower_np_binop(builder, target, target_type, args, linalg.sub)


@lower(operator.mul, types.Array, types.Array)
@lower(operator.mul, types.Array, types.Number)
@lower(operator.mul, types.Number, types.Array)
def operator_mul_array_lower(builder, target, args, kwargs):
    """Lower operator.mul for arrays by using linalg.mul"""
    from numba_cuda_mlir._mlir.dialects import linalg
    from numba_cuda_mlir.lowering_utilities.linalg_lowering import lower_np_binop

    target_type = builder.get_numba_type(target.name)
    lower_np_binop(builder, target, target_type, args, linalg.mul)


@lower(operator.imul, types.Array, types.Array)
@lower(operator.imul, types.Array, types.Number)
@lower(operator.imul, types.Number, types.Array)
def operator_imul_array_lower(builder, target, args, kwargs):
    """Lower operator.imul for arrays by using linalg.mul"""
    from numba_cuda_mlir._mlir.dialects import linalg
    from numba_cuda_mlir.lowering_utilities.linalg_lowering import lower_np_binop

    target_type = builder.get_numba_type(target.name)
    lower_np_binop(builder, target, target_type, args, linalg.mul)


@lower(operator.neg, types.Number)
def operator_neg_number_lower(builder, target, args, kwargs):
    """Lower operator.neg for numbers by using arith.negf"""
    value = builder.load_var(args[0])
    c0 = constant(0, value.type)
    result = c0 - value
    builder.store_var(target, result)


@lower(operator.eq, types.Number, types.NoneType)
@lower(operator.eq, types.NoneType, types.Number)
def operator_eq_none_lower(builder, target, args, kwargs):
    result = arith.constant(result=ir.IntegerType.get_signless(1), value=False)
    builder.store_var(target, result)


@lower(operator.ne, types.Number, types.NoneType)
@lower(operator.ne, types.NoneType, types.Number)
def operator_ne_none_lower(builder, target, args, kwargs):
    result = arith.constant(result=ir.IntegerType.get_signless(1), value=True)
    builder.store_var(target, result)


@lower(operator.eq, types.NoneType, types.NoneType)
def operator_eq_none_none_lower(builder, target, args, kwargs):
    result = arith.constant(result=ir.IntegerType.get_signless(1), value=True)
    builder.store_var(target, result)


@lower(operator.ne, types.NoneType, types.NoneType)
def operator_ne_none_none_lower(builder, target, args, kwargs):
    result = arith.constant(result=ir.IntegerType.get_signless(1), value=False)
    builder.store_var(target, result)


@lower(operator.is_, types.Number, types.NoneType)
@lower(operator.is_, types.NoneType, types.Number)
def operator_is_none_lower(builder, target, args, kwargs):
    result = arith.constant(result=ir.IntegerType.get_signless(1), value=False)
    builder.store_var(target, result)


@lower(operator.is_, types.BaseTuple, types.NoneType)
@lower(operator.is_, types.NoneType, types.BaseTuple)
def operator_is_tuple_none_lower(builder, target, args, kwargs):
    result = arith.constant(result=ir.IntegerType.get_signless(1), value=False)
    builder.store_var(target, result)


@lower(operator.is_, types.NoneType, types.NoneType)
def operator_is_none_none_lower(builder, target, args, kwargs):
    """Lower 'None is None' - always True."""
    result = arith.constant(result=ir.IntegerType.get_signless(1), value=True)
    builder.store_var(target, result)


@lower(operator.is_, types.Optional, types.NoneType)
@lower(operator.is_, types.NoneType, types.Optional)
def operator_is_optional_none_lower(builder, target, args, kwargs):
    """Lower 'optional_val is None' - check valid bit is false."""
    from numba_cuda_mlir.mlir.dialect_exts import llvm

    lhs_type = builder.get_numba_type(args[0].name)
    opt_arg = args[0] if isinstance(lhs_type, types.Optional) else args[1]
    opt_val = builder.load_var(opt_arg)
    i1 = ir.IntegerType.get_signless(1)
    valid = llvm.extractvalue(i1, opt_val, [1])
    result = arith.xori(valid, arith.constant(i1, 1))
    builder.store_var(target, result)


@lower(operator.is_, types.Boolean, types.Literal)
@lower(operator.is_, types.Literal, types.Boolean)
def operator_is_bool_literal_lower(builder, target, args, kwargs):
    """Lower 'x is True' or 'x is False' - compare values."""
    left = builder.load_var(args[0])
    right = builder.load_var(args[1])
    left_val = try_extract_constant(left)
    right_val = try_extract_constant(right)
    if left_val is not None and right_val is not None:
        result = arith.constant(result=ir.IntegerType.get_signless(1), value=left_val is right_val)
    elif left_val is not None:
        result = arith.cmpi(
            arith.CmpIPredicate.eq,
            arith.constant(result=ir.IntegerType.get_signless(1), value=left_val),
            right,
        )
    elif right_val is not None:
        result = arith.cmpi(
            arith.CmpIPredicate.eq,
            left,
            arith.constant(result=ir.IntegerType.get_signless(1), value=right_val),
        )
    else:
        result = arith.cmpi(arith.CmpIPredicate.eq, left, right)
    builder.store_var(target, result)


@lower(operator.is_not, types.Number, types.NoneType)
@lower(operator.is_not, types.NoneType, types.Number)
def operator_is_not_none_lower(builder, target, args, kwargs):
    result = arith.constant(result=ir.IntegerType.get_signless(1), value=True)
    builder.store_var(target, result)


@lower(operator.is_not, types.BaseTuple, types.NoneType)
@lower(operator.is_not, types.NoneType, types.BaseTuple)
def operator_is_not_tuple_none_lower(builder, target, args, kwargs):
    result = arith.constant(result=ir.IntegerType.get_signless(1), value=True)
    builder.store_var(target, result)


@lower(operator.is_not, types.NoneType, types.NoneType)
def operator_is_not_none_none_lower(builder, target, args, kwargs):
    """Lower 'None is not None' - always False."""
    result = arith.constant(result=ir.IntegerType.get_signless(1), value=False)
    builder.store_var(target, result)


@lower(operator.is_not, types.Optional, types.NoneType)
@lower(operator.is_not, types.NoneType, types.Optional)
def operator_is_not_optional_none_lower(builder, target, args, kwargs):
    """Lower 'optional_val is not None' - check valid bit is true."""
    from numba_cuda_mlir.mlir.dialect_exts import llvm

    lhs_type = builder.get_numba_type(args[0].name)
    opt_arg = args[0] if isinstance(lhs_type, types.Optional) else args[1]
    opt_val = builder.load_var(opt_arg)
    i1 = ir.IntegerType.get_signless(1)
    valid = llvm.extractvalue(i1, opt_val, [1])
    builder.store_var(target, valid)


@lower(operator.is_not, types.Boolean, types.Literal)
@lower(operator.is_not, types.Literal, types.Boolean)
def operator_is_not_bool_literal_lower(builder, target, args, kwargs):
    """Lower 'x is not True' or 'x is not False' - compare values."""
    left = builder.load_var(args[0])
    right = builder.load_var(args[1])
    left_val = try_extract_constant(left)
    right_val = try_extract_constant(right)
    if left_val is not None and right_val is not None:
        result = arith.constant(
            result=ir.IntegerType.get_signless(1), value=left_val is not right_val
        )
    elif left_val is not None:
        result = arith.cmpi(
            arith.CmpIPredicate.ne,
            arith.constant(result=ir.IntegerType.get_signless(1), value=left_val),
            right,
        )
    elif right_val is not None:
        result = arith.cmpi(
            arith.CmpIPredicate.ne,
            left,
            arith.constant(result=ir.IntegerType.get_signless(1), value=right_val),
        )
    else:
        result = arith.cmpi(arith.CmpIPredicate.ne, left, right)
    builder.store_var(target, result)


@lower(operator.pos, types.Number)
def operator_pos_number_lower(builder, target, args, kwargs):
    """Lower operator.neg for numbers by using arith.negf"""
    value = builder.load_var(args[0])
    builder.store_var(target, value)



# Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
# See https://llvm.org/LICENSE.txt for license information.
# SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
import sys
from ctypes import (
    POINTER,
    PYFUNCTYPE,
    Structure,
    _Pointer,
    c_char,
    c_char_p,
    c_int,
    c_ssize_t,
    c_uint,
    c_uint16,
    c_uint32,
    c_uint8,
    c_ulong,
    c_void_p,
    cast,
    pointer,
    py_object,
)
from typing import get_origin, _SpecialForm, _GenericAlias

_SelfPtr = object()


@_SpecialForm
def Self(self, parameters):
    raise TypeError(f"{self} is not subscriptable")


class _Ptr(_Pointer):
    def __new__(cls, *args, **kwargs):
        return pointer(*args, **kwargs)

    def __class_getitem__(cls, item):
        # For ptr[Self], return a special object
        if item is Self:
            return _SelfPtr

        # Get base of generic alias
        # noinspection PyUnresolvedReferences, PyProtectedMember
        if isinstance(item, _GenericAlias):
            item = get_origin(item)

        try:
            return POINTER(item)
        except TypeError as e:
            raise TypeError(f"{e} (During POINTER({item}))") from e


def address(obj) -> int:
    source = py_object(obj)
    addr = c_void_p.from_buffer(source).value
    if addr is None:
        raise ValueError("address: NULL object")  # pragma: no cover
    return addr


# https://github.com/python/cpython/blob/3.11/Include/object.h#L196-L227
unaryfunc = PYFUNCTYPE(py_object, py_object)
binaryfunc = PYFUNCTYPE(py_object, py_object, py_object)
ternaryfunc = PYFUNCTYPE(py_object, py_object, py_object, py_object)
inquiry = PYFUNCTYPE(c_int, py_object)
lenfunc = PYFUNCTYPE(c_ssize_t, py_object)
ssizeargfunc = PYFUNCTYPE(py_object, py_object, c_ssize_t)
ssizessizeargfunc = PYFUNCTYPE(py_object, py_object, c_ssize_t, c_ssize_t)
ssizeobjargproc = PYFUNCTYPE(c_int, py_object, c_ssize_t, py_object)
ssizessizeobjargproc = PYFUNCTYPE(c_int, py_object, c_ssize_t, c_ssize_t, py_object)
objobjargproc = PYFUNCTYPE(c_int, py_object, py_object, py_object)

objobjproc = PYFUNCTYPE(c_int, py_object, py_object)
visitproc = PYFUNCTYPE(c_int, py_object, c_void_p)
traverseproc = PYFUNCTYPE(c_int, py_object, visitproc, c_void_p)

freefunc = PYFUNCTYPE(None, c_void_p)
destructor = PYFUNCTYPE(None, py_object)
getattrfunc = PYFUNCTYPE(py_object, py_object, c_char_p)
getattrofunc = PYFUNCTYPE(py_object, py_object, py_object)
setattrfunc = PYFUNCTYPE(c_int, py_object, c_char_p, py_object)
setattrofunc = PYFUNCTYPE(c_int, py_object, py_object, py_object)
reprfunc = PYFUNCTYPE(py_object, py_object)
hashfunc = PYFUNCTYPE(c_ssize_t, py_object)
richcmpfunc = PYFUNCTYPE(py_object, py_object, py_object, c_int)
getiterfunc = PYFUNCTYPE(py_object, py_object)
iternextfunc = PYFUNCTYPE(py_object, py_object)

descrgetfunc = PYFUNCTYPE(py_object, py_object, py_object, py_object)
descrsetfunc = PYFUNCTYPE(c_int, py_object, py_object, py_object)
initproc = PYFUNCTYPE(c_int, py_object, py_object, py_object)
newfunc = PYFUNCTYPE(py_object, py_object, py_object, py_object)
allocfunc = PYFUNCTYPE(py_object, py_object, c_ssize_t)

# PyObject *(*vectorcallfunc)(PyObject *callable, PyObject *const *args, size_t nargsf, PyObject *kwnames)
vectorcallfunc = PYFUNCTYPE(py_object, py_object, _Ptr[py_object], c_ssize_t, py_object)
# PySendResult (*sendfunc)(PyObject *iter, PyObject *value, PyObject **result)
sendfunc = PYFUNCTYPE(c_int, py_object, py_object, _Ptr[py_object])


def _is_gil_enabled():
    try:
        return sys._is_gil_enabled()
    except:
        return True


if _is_gil_enabled():
    _py_object_fields = [
        ("ob_refcnt", c_ssize_t),
        ("ob_type", _Ptr[py_object]),
    ]
else:
    # https://github.com/python/cpython/blob/main/Include/object.h#L168
    _py_object_fields = [
        ("ob_tid", c_ssize_t),
        ("ob_flags", c_uint16),
        # https://github.com/python/cpython/blob/main/Include/cpython/pylock.h#L29
        ("ob_mutex", c_uint8),
        ("ob_gc_bits", c_uint8),
        ("ob_ref_local", c_uint32),
        ("ob_ref_shared", c_ssize_t),
        ("ob_type", _Ptr[py_object]),
    ]


class PyObject(Structure):
    _fields_ = _py_object_fields

    def as_ref(self) -> _Ptr[Self]:
        return pointer(self)  # type: ignore

    def into_object(self):
        py_obj = cast(self.as_ref(), py_object)
        return py_obj.value


# https://github.com/python/cpython/blob/3.11/Doc/includes/typestruct.h
class PyTypeObject(Structure):
    _fields_ = _py_object_fields + [
        ("ob_size", c_ssize_t),
        ("tp_name", c_char_p),
        ("tp_basicsize", c_ssize_t),
        ("tp_itemsize", c_ssize_t),
        ("tp_dealloc", destructor),
        ("tp_vectorcall_offset", c_ssize_t),
        ("tp_getattr", getattrfunc),
        ("tp_setattr", setattrfunc),
        # ("tp_as_async", _Ptr[PyAsyncMethods]),
        ("tp_as_async", _Ptr[PyObject]),
        ("tp_repr", reprfunc),
        # ("tp_as_number", _Ptr[PyNumberMethods]),
        ("tp_as_number", _Ptr[PyObject]),
        # ("tp_as_sequence", _Ptr[PySequenceMethods]),
        ("tp_as_sequence", _Ptr[PyObject]),
        # ("tp_as_mapping", _Ptr[PyMappingMethods]),
        ("tp_as_mapping", _Ptr[PyObject]),
        ("tp_hash", hashfunc),
        ("tp_call", ternaryfunc),
        ("tp_str", reprfunc),
        ("tp_getattro", getattrofunc),
        ("tp_setattro", setattrofunc),
        # ("tp_as_buffer", _Ptr[PyBufferProcs]),
        ("tp_as_buffer", _Ptr[PyObject]),
        ("tp_flags", c_ulong),
        ("tp_doc", c_char_p),
        ("tp_traverse", traverseproc),
        ("tp_clear", inquiry),
        ("tp_richcompare", richcmpfunc),
        ("tp_weaklistoffset", c_ssize_t),
        ("tp_iter", getiterfunc),
        ("tp_iternext", iternextfunc),
        ("tp_methods", _Ptr[PyObject]),
        ("tp_members", _Ptr[PyObject]),
        ("tp_getset", _Ptr[PyObject]),
        # ("tp_base", _Ptr[PyTypeObject]),
        ("tp_base", _Ptr[PyObject]),
        ("tp_dict", _Ptr[PyObject]),
        ("tp_descr_get", descrgetfunc),
        ("tp_descr_set", descrsetfunc),
        ("tp_dictoffset", c_ssize_t),
        ("tp_init", initproc),
        ("tp_alloc", allocfunc),
        ("tp_new", newfunc),
        ("tp_free", freefunc),
        ("tp_is_gc", inquiry),
        ("tp_bases", _Ptr[PyObject]),
        ("tp_mro", _Ptr[PyObject]),
        ("tp_cache", _Ptr[PyObject]),
        ("tp_subclasses", _Ptr[PyObject]),
        ("tp_weaklist", _Ptr[PyObject]),
        ("tp_del", destructor),
        ("tp_version_tag", c_uint),
        ("tp_finalize", descrsetfunc),
        ("tp_vectorcall", vectorcallfunc),
        ("tp_watched", c_char),
    ]

    @classmethod
    def from_object(cls, obj) -> Self:
        return cls.from_address(address(obj))


# https://github.com/python/cpython/blob/9648eed33f5fca0fcb8802fe0be8d35907bc33e3/Objects/typevarobject.c#L19-L31
class PyTypeVarObject(Structure):
    _fields_ = _py_object_fields + [
        ("ob_size", c_ssize_t),
        ("name", _Ptr[PyObject]),
        ("bound", _Ptr[PyObject]),
        ("evaluate_bound", _Ptr[PyObject]),
        ("constraints", _Ptr[PyObject]),
        ("evaluate_constraints", _Ptr[PyObject]),
        ("default_value", _Ptr[PyObject]),
        ("evaluate_default", _Ptr[PyObject]),
    ]

    @classmethod
    def from_object(cls, obj) -> Self:
        return cls.from_address(address(obj))




# Autogenerated by mlir-tblgen; don't manually edit.

from ._ods_common import _cext as _ods_cext
from ._ods_common import (
    equally_sized_accessor as _ods_equally_sized_accessor,
    get_default_loc_context as _ods_get_default_loc_context,
    get_op_results_or_values as _get_op_results_or_values,
    segmented_accessor as _ods_segmented_accessor,
)
_ods_ir = _ods_cext.ir
_ods_cext.globals.register_traceback_file_exclusion(__file__)

import builtins
from typing import Any as _Any, Sequence as _Sequence, Union as _Union, Optional as _Optional
import sys as _sys
if _sys.version_info >= (3, 12):
  from collections.abc import Buffer as _Buffer  # pytype: disable=not-supported-yet
else:
  try:
    from typing_extensions import Buffer as _Buffer
  except ImportError:
    _Buffer = _Any


from ._transform_ops_gen import _Dialect

@_ods_cext.register_operation(_Dialect)
class BufferLoopHoistingOp(_ods_ir.OpView):
  r"""
  Hoist buffer allocations ("memref.alloc" and "memref.alloca") from loops
  within the targeted op. This transform assumes that there are no buffer
  deallocation ops in the IR.
  
  This transform reads the `target` handle and modifies the payload.
  """

  OPERATION_NAME = "transform.bufferization.buffer_loop_hoisting"

  _ODS_REGIONS = (0, True)

  def __init__(self, target: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(target)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operation.operands[0]

@_ods_cext.register_op_adaptor(BufferLoopHoistingOp)
class BufferLoopHoistingOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "transform.bufferization.buffer_loop_hoisting"

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operands[0]

def bufferization_buffer_loop_hoisting(target: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> BufferLoopHoistingOp:
  return BufferLoopHoistingOp(target=target, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class EliminateEmptyTensorsOp(_ods_ir.OpView):
  r"""
  Try to eliminate all `tensor.empty` ops within the targeted op by replacing
  them with another destination tensor.
  
  "tensor.empty" ops cannot be bufferized. They can either be converted to
  "bufferization.alloc_tensor" or replaced with another tensor (via this
  transform). "tensor.empty" does not specify the contents of the returned
  tensor so their results can be replaced with arbitrary tensor values as long
  as the dimensions match.
  
  This transformation looks for subset ops that insert a tensor that
  originates from a "tensor.empty" (as per the reverse use-def chain). Such
  "tensor.empty" ops are replaced with the destination subset.
  
  Example:
  
  ```
  %0 = tensor.empty() : tensor<5xf32>
  %1 = linalg.fill ... outs(%0)
  %2 = tensor.insert_slice %1 into %t[1][5][1]
  ```
  
  Is rewritten with:
  ```
  %0 = tensor.extract_slice %t[1][5][1]
  %1 = linalg.fill ... outs(%0)
  %2 = tensor.insert_slice %1 into %t[1][5][1]
  ```
  
  In the above example, the subset op is "tensor.insert_slice". When tracing
  back the reverse use-def chain of a the source, we end up at a
  "tensor.empty" op.
  
  The above example can bufferize without an allocation (in the absence of
  other conflicts) because there is no longer a `tensor.empty` op.
  
  See `-eliminate-empty-tensors` for more details.
  
  #### Return modes
  
  This transform reads the target handle and modifies the payload. It does
  not produce any handle.
  """

  OPERATION_NAME = "transform.bufferization.eliminate_empty_tensors"

  _ODS_REGIONS = (0, True)

  def __init__(self, target: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(target)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operation.operands[0]

@_ods_cext.register_op_adaptor(EliminateEmptyTensorsOp)
class EliminateEmptyTensorsOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "transform.bufferization.eliminate_empty_tensors"

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operands[0]

def bufferization_eliminate_empty_tensors(target: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> EliminateEmptyTensorsOp:
  return EliminateEmptyTensorsOp(target=target, loc=loc, ip=ip)

@_ods_cext.register_operation(_Dialect)
class EmptyTensorToAllocTensorOp(_ods_ir.OpView):
  r"""
  Replace a tensor.empty with a bufferization.tensor_alloc.
  
  #### Return modes
  
  This operation consumes the `target` handle and produces the `transformed`
  handle. `target` is expected to be a `tensor.empty` operation. The transform
  always succeeds.
  """

  OPERATION_NAME = "transform.bufferization.empty_tensor_to_alloc_tensor"

  _ODS_REGIONS = (0, True)

  def __init__(self, transformed: _ods_ir.Type, target: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(target)
    _ods_context = _ods_get_default_loc_context(loc)
    results = []
    results.append(transformed)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def transformed(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(EmptyTensorToAllocTensorOp)
class EmptyTensorToAllocTensorOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "transform.bufferization.empty_tensor_to_alloc_tensor"

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operands[0]

def bufferization_empty_tensor_to_alloc_tensor(transformed: _ods_ir.Type, target: _ods_ir.Value, *, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return EmptyTensorToAllocTensorOp(transformed=transformed, target=target, loc=loc, ip=ip).result

@_ods_cext.register_operation(_Dialect)
class OneShotBufferizeOp(_ods_ir.OpView):
  r"""
  Indicates that the given `target` op should be bufferized with One-Shot
  Bufferize. The bufferization can be configured with various attributes that
  corresponding to options in `BufferizationOptions` and the
  `one-shot-bufferize` pass. More information can be found in the pass
  documentation.
  
  The targeted ops must be modules or functions. This is because there is
  always a single, bufferized replacement op for such targets.
  
  Note: Only ops that implement `BufferizableOpInterface` are bufferized. All
  other ops are ignored if `allow_unknown_ops`. If `allow_unknown_ops` is
  unset, this transform fails when an unknown/non-bufferizable op is found.
  Many ops implement `BufferizableOpInterface` via an external model. These
  external models must be registered when applying this transform op;
  otherwise, said ops would be considered non-bufferizable.
  
  #### Return modes
  
  This operation consumes the `target` handle and produces the `transformed`
  handle.
  """

  OPERATION_NAME = "transform.bufferization.one_shot_bufferize"

  _ODS_REGIONS = (0, True)

  def __init__(self, transformed: _ods_ir.Type, target: _ods_ir.Value, *, function_boundary_type_conversion: _Optional[_Union[_Any, _ods_ir.Attribute]] = None, allow_return_allocs_from_loops: _Optional[_Union[bool, _ods_ir.BoolAttr]] = None, allow_unknown_ops: _Optional[_Union[bool, _ods_ir.BoolAttr]] = None, bufferize_function_boundaries: _Optional[_Union[bool, _ods_ir.BoolAttr]] = None, dump_alias_sets: _Optional[_Union[bool, _ods_ir.BoolAttr]] = None, test_analysis_only: _Optional[_Union[bool, _ods_ir.BoolAttr]] = None, print_conflicts: _Optional[_Union[bool, _ods_ir.BoolAttr]] = None, check_parallel_regions: _Optional[_Union[bool, _ods_ir.BoolAttr]] = None, memcpy_op: _Optional[_Union[str, _ods_ir.StringAttr]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None):
    operands = []
    attributes = {}
    regions = None
    operands.append(target)
    _ods_context = _ods_get_default_loc_context(loc)
    if function_boundary_type_conversion is not None: attributes["function_boundary_type_conversion"] = (function_boundary_type_conversion if (
        isinstance(function_boundary_type_conversion, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('LayoutMapOption')) else
          _ods_ir.AttrBuilder.get('LayoutMapOption')(function_boundary_type_conversion, context=_ods_context))
    if allow_return_allocs_from_loops is not None: attributes["allow_return_allocs_from_loops"] = (allow_return_allocs_from_loops if (
        isinstance(allow_return_allocs_from_loops, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('BoolAttr')) else
          _ods_ir.AttrBuilder.get('BoolAttr')(allow_return_allocs_from_loops, context=_ods_context))
    if allow_unknown_ops is not None: attributes["allow_unknown_ops"] = (allow_unknown_ops if (
        isinstance(allow_unknown_ops, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('BoolAttr')) else
          _ods_ir.AttrBuilder.get('BoolAttr')(allow_unknown_ops, context=_ods_context))
    if bufferize_function_boundaries is not None: attributes["bufferize_function_boundaries"] = (bufferize_function_boundaries if (
        isinstance(bufferize_function_boundaries, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('BoolAttr')) else
          _ods_ir.AttrBuilder.get('BoolAttr')(bufferize_function_boundaries, context=_ods_context))
    if dump_alias_sets is not None: attributes["dump_alias_sets"] = (dump_alias_sets if (
        isinstance(dump_alias_sets, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('BoolAttr')) else
          _ods_ir.AttrBuilder.get('BoolAttr')(dump_alias_sets, context=_ods_context))
    if test_analysis_only is not None: attributes["test_analysis_only"] = (test_analysis_only if (
        isinstance(test_analysis_only, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('BoolAttr')) else
          _ods_ir.AttrBuilder.get('BoolAttr')(test_analysis_only, context=_ods_context))
    if print_conflicts is not None: attributes["print_conflicts"] = (print_conflicts if (
        isinstance(print_conflicts, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('BoolAttr')) else
          _ods_ir.AttrBuilder.get('BoolAttr')(print_conflicts, context=_ods_context))
    if check_parallel_regions is not None: attributes["check_parallel_regions"] = (check_parallel_regions if (
        isinstance(check_parallel_regions, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('BoolAttr')) else
          _ods_ir.AttrBuilder.get('BoolAttr')(check_parallel_regions, context=_ods_context))
    if memcpy_op is not None: attributes["memcpy_op"] = (memcpy_op if (
        isinstance(memcpy_op, _ods_ir.Attribute) or
        not _ods_ir.AttrBuilder.contains('StrAttr')) else
          _ods_ir.AttrBuilder.get('StrAttr')(memcpy_op, context=_ods_context))
    results = []
    results.append(transformed)
    _ods_successors = None
    super().__init__(self.OPERATION_NAME, self._ODS_REGIONS, self._ODS_OPERAND_SEGMENTS, self._ODS_RESULT_SEGMENTS, attributes=attributes, results=results, operands=operands, successors=_ods_successors, regions=regions, loc=loc, ip=ip)

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operation.operands[0]

  @builtins.property
  def function_boundary_type_conversion(self) -> _Optional[_ods_ir.Attribute]:
    if "function_boundary_type_conversion" not in self.operation.attributes:
      return None
    return self.operation.attributes["function_boundary_type_conversion"]

  @function_boundary_type_conversion.setter
  def function_boundary_type_conversion(self, value: _Optional[_ods_ir.Attribute]):
    if value is not None:
      self.operation.attributes["function_boundary_type_conversion"] = value
    elif "function_boundary_type_conversion" in self.operation.attributes:
      del self.operation.attributes["function_boundary_type_conversion"]

  @function_boundary_type_conversion.deleter
  def function_boundary_type_conversion(self):
    del self.operation.attributes["function_boundary_type_conversion"]

  @builtins.property
  def allow_return_allocs_from_loops(self) -> _ods_ir.BoolAttr:
    return self.operation.attributes["allow_return_allocs_from_loops"]

  @allow_return_allocs_from_loops.setter
  def allow_return_allocs_from_loops(self, value: _ods_ir.BoolAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["allow_return_allocs_from_loops"] = value

  @builtins.property
  def allow_unknown_ops(self) -> _ods_ir.BoolAttr:
    return self.operation.attributes["allow_unknown_ops"]

  @allow_unknown_ops.setter
  def allow_unknown_ops(self, value: _ods_ir.BoolAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["allow_unknown_ops"] = value

  @builtins.property
  def bufferize_function_boundaries(self) -> _ods_ir.BoolAttr:
    return self.operation.attributes["bufferize_function_boundaries"]

  @bufferize_function_boundaries.setter
  def bufferize_function_boundaries(self, value: _ods_ir.BoolAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["bufferize_function_boundaries"] = value

  @builtins.property
  def dump_alias_sets(self) -> _ods_ir.BoolAttr:
    return self.operation.attributes["dump_alias_sets"]

  @dump_alias_sets.setter
  def dump_alias_sets(self, value: _ods_ir.BoolAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["dump_alias_sets"] = value

  @builtins.property
  def test_analysis_only(self) -> _ods_ir.BoolAttr:
    return self.operation.attributes["test_analysis_only"]

  @test_analysis_only.setter
  def test_analysis_only(self, value: _ods_ir.BoolAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["test_analysis_only"] = value

  @builtins.property
  def print_conflicts(self) -> _ods_ir.BoolAttr:
    return self.operation.attributes["print_conflicts"]

  @print_conflicts.setter
  def print_conflicts(self, value: _ods_ir.BoolAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["print_conflicts"] = value

  @builtins.property
  def check_parallel_regions(self) -> _ods_ir.BoolAttr:
    return self.operation.attributes["check_parallel_regions"]

  @check_parallel_regions.setter
  def check_parallel_regions(self, value: _ods_ir.BoolAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["check_parallel_regions"] = value

  @builtins.property
  def memcpy_op(self) -> _ods_ir.StringAttr:
    return self.operation.attributes["memcpy_op"]

  @memcpy_op.setter
  def memcpy_op(self, value: _ods_ir.StringAttr):
    if value is None:
      raise ValueError("'None' not allowed as value for mandatory attributes")
    self.operation.attributes["memcpy_op"] = value

  @builtins.property
  def transformed(self) -> _ods_ir.OpResult:
    return self.operation.results[0]

@_ods_cext.register_op_adaptor(OneShotBufferizeOp)
class OneShotBufferizeOpAdaptor(_ods_ir.OpAdaptor):
  OPERATION_NAME = "transform.bufferization.one_shot_bufferize"

  @builtins.property
  def target(self) -> _ods_ir.Value:
    return self.operands[0]

  @builtins.property
  def function_boundary_type_conversion(self) -> _Optional[_ods_ir.Attribute]:
    if "function_boundary_type_conversion" not in self.attributes:
      return None
    return self.attributes["function_boundary_type_conversion"]

  @builtins.property
  def allow_return_allocs_from_loops(self) -> _ods_ir.BoolAttr:
    return self.attributes["allow_return_allocs_from_loops"]

  @builtins.property
  def allow_unknown_ops(self) -> _ods_ir.BoolAttr:
    return self.attributes["allow_unknown_ops"]

  @builtins.property
  def bufferize_function_boundaries(self) -> _ods_ir.BoolAttr:
    return self.attributes["bufferize_function_boundaries"]

  @builtins.property
  def dump_alias_sets(self) -> _ods_ir.BoolAttr:
    return self.attributes["dump_alias_sets"]

  @builtins.property
  def test_analysis_only(self) -> _ods_ir.BoolAttr:
    return self.attributes["test_analysis_only"]

  @builtins.property
  def print_conflicts(self) -> _ods_ir.BoolAttr:
    return self.attributes["print_conflicts"]

  @builtins.property
  def check_parallel_regions(self) -> _ods_ir.BoolAttr:
    return self.attributes["check_parallel_regions"]

  @builtins.property
  def memcpy_op(self) -> _ods_ir.StringAttr:
    return self.attributes["memcpy_op"]

def bufferization_one_shot_bufferize(transformed: _ods_ir.Type, target: _ods_ir.Value, *, function_boundary_type_conversion: _Optional[_Union[_Any, _ods_ir.Attribute]] = None, allow_return_allocs_from_loops: _Optional[_Union[bool, _ods_ir.BoolAttr]] = None, allow_unknown_ops: _Optional[_Union[bool, _ods_ir.BoolAttr]] = None, bufferize_function_boundaries: _Optional[_Union[bool, _ods_ir.BoolAttr]] = None, dump_alias_sets: _Optional[_Union[bool, _ods_ir.BoolAttr]] = None, test_analysis_only: _Optional[_Union[bool, _ods_ir.BoolAttr]] = None, print_conflicts: _Optional[_Union[bool, _ods_ir.BoolAttr]] = None, check_parallel_regions: _Optional[_Union[bool, _ods_ir.BoolAttr]] = None, memcpy_op: _Optional[_Union[str, _ods_ir.StringAttr]] = None, loc: _Optional[_ods_ir.Location] = None, ip: _Optional[_ods_ir.InsertionPoint] = None) -> _ods_ir.OpResult:
  return OneShotBufferizeOp(transformed=transformed, target=target, function_boundary_type_conversion=function_boundary_type_conversion, allow_return_allocs_from_loops=allow_return_allocs_from_loops, allow_unknown_ops=allow_unknown_ops, bufferize_function_boundaries=bufferize_function_boundaries, dump_alias_sets=dump_alias_sets, test_analysis_only=test_analysis_only, print_conflicts=print_conflicts, check_parallel_regions=check_parallel_regions, memcpy_op=memcpy_op, loc=loc, ip=ip).result
